from shiny import App, Inputs, Outputs, Session, ui

app_ui = ui.page_fluid(
    ui.input_checkbox("show", "Show radio buttons", False),
    ui.panel_conditional(
        "input.show", ui.input_radio_buttons("radio", "Choose ", ["slider", "select"])
    ),
    ui.panel_conditional(
        "input.show && input.radio === 'slider'",
        ui.input_slider("slider", None, min=0, max=100, value=50),
    ),
    ui.panel_conditional(
        "input.show && input.radio === 'select'",
        ui.input_select("select", None, ["A", "B", "C"]),
    ),
)


def server(input: Inputs, output: Outputs, session: Session):
    pass


app = App(app_ui, server)



from palmerpenguins import load_penguins_raw  # pyright: ignore[reportMissingTypeStubs]

from shiny import App, Inputs, Outputs, Session, render, ui

app_ui = ui.page_fluid(
    ui.h2("Palmer Penguins"),
    ui.output_data_frame("penguins_df"),
    ui.output_code("selected_row_count"),
)


def server(input: Inputs, output: Outputs, session: Session) -> None:
    @render.data_frame
    def penguins_df():
        return render.DataGrid(
            data=load_penguins_raw(),  # pyright: ignore[reportUnknownArgumentType]
            filters=True,
            selection_mode="rows",
        )

    @render.code
    def selected_row_count():
        # grid_selected_data = grid_selected.data()
        return str(
            penguins_df.data_view(selected=True)
        )  # pyright: ignore[reportUnknownMemberType]


app = App(app_ui, server)



"""Tests for `shiny.reactive`."""

import asyncio
from typing import List

import pytest

from shiny import App, render, req, ui
from shiny._connection import MockConnection
from shiny.reactive import Value, calc, effect, event, flush, invalidate_later, isolate
from shiny.reactive._core import ReactiveWarning
from shiny.types import ActionButtonValue, SilentException

from .mocktime import MockTime


@pytest.mark.asyncio
async def test_flush_runs_newly_invalidated():
    """
    Make sure that a flush will also run any calcs that were invalidated during the
    flush.
    """

    v1 = Value(1)
    v2 = Value(2)

    v2_result = None

    # In practice, on the first flush, Effects run in the order that they were created.
    # Our test checks that o2 runs _after_ o1.
    @effect()
    def o2():
        nonlocal v2_result
        v2_result = v2()

    @effect()
    def o1():
        v2.set(v1())

    await flush()
    assert v2_result == 1
    assert o2._exec_count == 2
    assert o1._exec_count == 1


@pytest.mark.asyncio
async def test_flush_runs_newly_invalidated_async():
    """
    Make sure that a flush will also run any calcs that were invalidated during the
    flush. (Same as previous test, but async.)
    """

    v1 = Value(1)
    v2 = Value(2)

    v2_result = None

    # In practice, on the first flush, Effects run in the order that they were
    # created. Our test checks that o2 runs _after_ o1.
    @effect()
    async def o2():
        nonlocal v2_result
        v2_result = v2()

    @effect()
    async def o1():
        v2.set(v1())

    await flush()
    assert v2_result == 1
    assert o2._exec_count == 2
    assert o1._exec_count == 1


# ======================================================================
# Setting Value to same value doesn't invalidate downstream
# ======================================================================
@pytest.mark.asyncio
async def test_reactive_value_same_no_invalidate():
    v = Value(1)

    @effect()
    def o():
        v()

    await flush()
    assert o._exec_count == 1

    v.set(1)
    await flush()
    assert o._exec_count == 1


# ======================================================================
# Intializing reactive.Value to MISSING, and unsetting
# ======================================================================
@pytest.mark.asyncio
async def test_reactive_value_unset():
    v = Value[int]()

    with isolate():
        assert v.is_set() is False
        with pytest.raises(SilentException):
            v()

    val: int = 0

    @effect()
    def o():
        nonlocal val
        val = v()

    await flush()
    assert o._exec_count == 1
    assert val == 0

    v.set(1)
    await flush()
    assert o._exec_count == 2
    assert val == 1
    with isolate():
        assert v.is_set() is True

    v.unset()
    await flush()
    assert o._exec_count == 3
    assert val == 1
    with isolate():
        assert v.is_set() is False
        with pytest.raises(SilentException):
            v()


# ======================================================================
# reactive.Value.is_set() invalidates dependents only when set state changes
# ======================================================================
@pytest.mark.asyncio
async def test_reactive_value_is_set():
    v = Value[int]()
    v_is_set: bool = False

    @effect()
    def o():
        nonlocal v_is_set
        v_is_set = v.is_set()

    await flush()
    assert o._exec_count == 1
    assert v_is_set is False

    v.set(1)
    await flush()
    assert o._exec_count == 2
    assert v_is_set is True

    v.set(2)
    await flush()
    assert o._exec_count == 2
    assert v_is_set is True

    v.unset()
    await flush()
    assert o._exec_count == 3
    assert v_is_set is False

    v.unset()
    await flush()
    assert o._exec_count == 3
    assert v_is_set is False

    v.set(1)
    await flush()
    assert o._exec_count == 4
    assert v_is_set is True


# ======================================================================
# Recursive calls to calcs
# ======================================================================
@pytest.mark.asyncio
async def test_recursive_calc():
    v = Value(5)

    @calc()
    def r():
        if v() == 0:
            return 0
        v.set(v() - 1)
        r()

    @effect()
    def o():
        r()

    await flush()
    assert o._exec_count == 2
    assert r._exec_count == 6
    with isolate():
        assert v() == 0


@pytest.mark.asyncio
async def test_recursive_async_calc():
    v = Value(5)

    @calc()
    async def r() -> int:
        if v() == 0:
            return 0
        v.set(v() - 1)
        return await r()

    @effect()
    async def o():
        await r()

    await flush()
    assert o._exec_count == 2
    assert r._exec_count == 6
    with isolate():
        assert v() == 0


# ======================================================================
# async
# ======================================================================


@pytest.mark.asyncio
async def test_async_sequential():
    x: Value[int] = Value(1)
    results: list[int] = []
    exec_order: list[str] = []

    async def react_chain(n: int):
        @calc()
        async def r():
            nonlocal exec_order
            exec_order.append(f"r{n}-1")
            await asyncio.sleep(0)
            exec_order.append(f"r{n}-2")
            return x() + 10

        @effect()
        async def _():
            nonlocal exec_order
            exec_order.append(f"o{n}-1")
            await asyncio.sleep(0)
            exec_order.append(f"o{n}-2")
            val = await r()
            exec_order.append(f"o{n}-3")
            results.append(val + n * 100)

    await asyncio.gather(react_chain(1), react_chain(2))
    await flush()
    x.set(5)
    await flush()

    assert results == [111, 211, 115, 215]

    # This is the order of execution if the async effects are run sequentially. The
    # `asyncio.sleep(0)` still yields control, but since there are no other effects
    # scheduled, it will simply resume at the same point.
    # fmt: off
    assert exec_order == [
        'o1-1', 'o1-2', 'r1-1', 'r1-2', 'o1-3',
        'o2-1', 'o2-2', 'r2-1', 'r2-2', 'o2-3',
        'o1-1', 'o1-2', 'r1-1', 'r1-2', 'o1-3',
        'o2-1', 'o2-2', 'r2-1', 'r2-2', 'o2-3'
    ]
    # fmt: on


# ======================================================================
# isolate()
# ======================================================================
@pytest.mark.asyncio
async def test_isolate_basic_without_context():
    # isolate() works with calc and Value; allows executing without a reactive context.
    v = Value(1)

    @calc()
    def r():
        return v() + 10

    def get_r():
        return r()

    with isolate():
        assert v() == 1
        assert (lambda: v())() == 1
        assert r() == 11
        assert (lambda: r())() == 11
        assert get_r() == 11


@pytest.mark.asyncio
async def test_isolate_prevents_dependency():
    v = Value(1)

    @calc()
    def r():
        return v() + 10

    v_dep = Value(1)  # Use this only for invalidating the effect
    o_val = None

    @effect()
    def o():
        nonlocal o_val
        v_dep()
        with isolate():
            o_val = r()

    await flush()
    assert o_val == 11

    # Changing v() shouldn't invalidate o
    v.set(2)
    await flush()
    assert o_val == 11
    assert o._exec_count == 1

    # v_dep() should invalidate the effect
    v_dep.set(2)
    await flush()
    assert o_val == 12
    assert o._exec_count == 2


# ======================================================================
# async isolate
# ======================================================================
@pytest.mark.asyncio
async def test_isolate_async_basic_value():
    async def f():
        return 123

    with isolate():
        assert await f() == 123


@pytest.mark.asyncio
async def test_isolate_async_basic_without_context():
    # async isolate works with calc and Value; allows executing without a reactive
    # context.
    v = Value(1)

    @calc()
    async def r():
        return v() + 10

    async def get_r():
        return await r()

    with isolate():
        assert await r() == 11
        assert await get_r() == 11


@pytest.mark.asyncio
async def test_isolate_async_prevents_dependency():
    v = Value(1)

    @calc()
    async def r():
        return v() + 10

    v_dep = Value(1)  # Use this only for invalidating the effect
    o_val = None

    @effect()
    async def o():
        nonlocal o_val
        v_dep()
        with isolate():
            o_val = await r()

    await flush()
    assert o_val == 11

    # Changing v() shouldn't invalidate o
    v.set(2)
    await flush()
    assert o_val == 11
    assert o._exec_count == 1

    # v_dep() should invalidate the effect
    v_dep.set(2)
    await flush()
    assert o_val == 12
    assert o._exec_count == 2


# ======================================================================
# Priority for effects
# ======================================================================
@pytest.mark.asyncio
async def test_effect_priority():
    v = Value(1)
    results: list[int] = []

    @effect(priority=1)
    def o1():
        nonlocal results
        v()
        results.append(1)

    @effect(priority=2)
    def o2():
        nonlocal results
        v()
        results.append(2)

    @effect(priority=1)
    def o3():
        nonlocal results
        v()
        results.append(3)

    await flush()
    assert results == [2, 1, 3]

    # Add another effect with priority 2. Only this one will run (until we
    # invalidate others by changing v).
    @effect(priority=2)
    def o4():
        nonlocal results
        v()
        results.append(4)

    results.clear()
    await flush()
    assert results == [4]

    # Change v and run again, to make sure results are stable
    results.clear()
    v.set(2)
    await flush()
    assert results == [2, 4, 1, 3]

    results.clear()
    v.set(3)
    await flush()
    assert results == [2, 4, 1, 3]


# Same as previous, but with async
@pytest.mark.asyncio
async def test_async_effect_priority():
    v = Value(1)
    results: list[int] = []

    @effect(priority=1)
    async def o1():
        nonlocal results
        v()
        results.append(1)

    @effect(priority=2)
    async def o2():
        nonlocal results
        v()
        results.append(2)

    @effect(priority=1)
    async def o3():
        nonlocal results
        v()
        results.append(3)

    await flush()
    assert results == [2, 1, 3]

    # Add another effect with priority 2. Only this one will run (until we
    # invalidate others by changing v).
    @effect(priority=2)
    async def o4():
        nonlocal results
        v()
        results.append(4)

    results.clear()
    await flush()
    assert results == [4]

    # Change v and run again, to make sure results are stable
    results.clear()
    v.set(2)
    await flush()
    assert results == [2, 4, 1, 3]

    results.clear()
    v.set(3)
    await flush()
    assert results == [2, 4, 1, 3]


# ======================================================================
# Destroying effects
# ======================================================================
@pytest.mark.asyncio
async def test_effect_destroy():
    v = Value(1)
    results: list[int] = []

    @effect()
    def o1():
        nonlocal results
        v()
        results.append(1)

    await flush()
    assert results == [1]

    v.set(2)
    o1.destroy()
    await flush()
    assert results == [1]

    # Same as above, but destroy before running first time
    v = Value(1)
    results: list[int] = []

    @effect()
    def o2():
        nonlocal results
        v()
        results.append(1)

    o2.destroy()
    await flush()
    assert results == []


# ======================================================================
# Error handling
# ======================================================================
@pytest.mark.asyncio
async def test_error_handling():
    vals: List[str] = []

    @effect()
    def _():
        vals.append("o1")

    @effect()
    def _():
        vals.append("o2-1")
        raise Exception("Error here!")
        vals.append("o2-2")

    @effect()
    def _():
        vals.append("o3")

    # Error in effect should get converted to warning.
    with pytest.warns(ReactiveWarning):
        await flush()
    # All effects should have executed.
    assert vals == ["o1", "o2-1", "o3"]

    vals: List[str] = []

    @calc()
    def r():
        vals.append("r")
        raise Exception("Error here!")

    @effect()
    def _():
        vals.append("o1-1")
        r()
        vals.append("o1-2")

    @effect()
    def _():
        vals.append("o2")

    # Error in effect should get converted to warning.
    with pytest.warns(ReactiveWarning):
        await flush()
    assert vals == ["o1-1", "r", "o2"]


@pytest.mark.asyncio
async def test_calc_error_rethrow():
    # Make sure calcs re-throw errors.
    vals: List[str] = []
    v = Value(1)

    @calc()
    def r():
        vals.append("r")
        raise Exception("Error here!")

    @effect()
    def _():
        v()
        vals.append("o1-1")
        r()
        vals.append("o1-2")

    @effect()
    def _():
        v()
        vals.append("o2-2")
        r()
        vals.append("o2-2")

    with pytest.warns(ReactiveWarning):
        await flush()
    assert vals == ["o1-1", "r", "o2-2"]

    v.set(2)
    with pytest.warns(ReactiveWarning):
        await flush()
    assert vals == ["o1-1", "r", "o2-2", "o1-1", "o2-2"]


# ======================================================================
# Invalidating dependents
# ======================================================================
# For https://github.com/posit-dev/py-shiny/issues/26
@pytest.mark.asyncio
async def test_dependent_invalidation():
    trigger = Value(0)
    v = Value(0)
    error_occurred = False

    @effect()
    def _():
        trigger()

        try:
            with isolate():
                r()
                val = v()
                v.set(val + 1)
        except Exception:
            nonlocal error_occurred
            error_occurred = True

    @effect()
    def _():
        r()

    @calc()
    def r():
        return v()

    await flush()
    trigger.set(1)
    await flush()

    with isolate():
        val = v()

    assert val == 2
    assert error_occurred is False


# ------------------------------------------------------------
# req() pauses execution in @effect() and @calc()
# ------------------------------------------------------------
@pytest.mark.asyncio
async def test_req():
    n_times = 0

    @effect()
    def _():
        req(False)
        nonlocal n_times
        n_times += 1

    await flush()
    assert n_times == 0

    @effect()
    def _():
        req(True)
        nonlocal n_times
        n_times += 1

    await flush()
    assert n_times == 1

    @calc()
    def r():
        req(False)
        return 1

    val = None

    @effect()
    def _():
        nonlocal val
        val = r()

    await flush()
    assert val is None

    @calc()
    def r2():
        req(True)
        return 1

    @effect()
    def _():
        nonlocal val
        val = r2()

    await flush()
    assert val == 1


@pytest.mark.asyncio
async def test_invalidate_later():
    mock_time = MockTime()
    with mock_time():

        @effect()
        def obs1():
            invalidate_later(1)

        # Initial run happens immediately
        await flush()
        assert obs1._exec_count == 1

        # If not enough time passes, no re-execution occurs
        await mock_time.advance_time(0.5)
        assert obs1._exec_count == 1

        # Re-execute after one second (total)
        await mock_time.advance_time(0.5)
        assert obs1._exec_count == 2

        # Make sure it happens repeatedly
        await mock_time.advance_time(10)
        assert obs1._exec_count == 12

        # After destruction, no more re-executions
        obs1.destroy()
        await mock_time.advance_time(10)
        assert obs1._exec_count == 12


@pytest.mark.asyncio
async def test_invalidate_later_invalidation():
    mock_time = MockTime()
    with mock_time():
        rv = Value(0)

        @effect()
        def obs1():
            if rv() == 0:
                invalidate_later(1)

        await flush()
        assert obs1._exec_count == 1

        # Change rv, triggering invalidation of obs1. The expected behavior is that
        # the invalidation causes the invalidate_later call to be cancelled.
        rv.set(1)
        await flush()
        assert obs1._exec_count == 2

        # Advance time to long after the invalidate_later would have fired, and ensure
        # nothing happens.
        await mock_time.advance_time(10)
        assert obs1._exec_count == 2
        await flush()
        assert obs1._exec_count == 2


@pytest.mark.asyncio
async def test_mock_time():
    mock_time = MockTime()

    with mock_time():
        results: List[str] = []

        async def add_result_later(delay: float, msg: str):
            await asyncio.sleep(delay)
            results.append(msg)

        asyncio.create_task(add_result_later(10, "a"))
        asyncio.create_task(add_result_later(5, "b"))
        task_c = asyncio.create_task(add_result_later(15, "c"))

        await mock_time.advance_time(11)

        assert results == ["b", "a"]
        # Prevent asyncio from complaining about a pending task not being complete
        task_c.cancel()


# ------------------------------------------------------------
# @reactive.event() works as expected
# ------------------------------------------------------------
@pytest.mark.asyncio
async def test_event_decorator():
    n_times = 0

    # By default, runs every time that event expression is _not_ None (ignore_none=True)
    @effect()
    @event(lambda: None, lambda: ActionButtonValue(0))
    def _():
        nonlocal n_times
        n_times += 1

    await flush()
    assert n_times == 0

    # Unless ignore_none=False
    @effect()
    @event(lambda: None, lambda: ActionButtonValue(0), ignore_none=False)
    def _():
        nonlocal n_times
        n_times += 1

    await flush()
    assert n_times == 1

    # Or if one of the args is not None
    @effect()
    @event(lambda: None, lambda: ActionButtonValue(0), lambda: True)
    def _():
        nonlocal n_times
        n_times += 1

    await flush()
    assert n_times == 2

    # Is invalidated properly by reactive values
    v = Value(1)

    @effect()
    @event(v)
    def _():
        nonlocal n_times
        n_times += 1

    await flush()
    assert n_times == 3

    v.set(1)
    await flush()
    assert n_times == 3

    v.set(2)
    await flush()
    assert n_times == 4

    # Doesn't run on init
    v = Value(1)

    @effect()
    @event(v, ignore_init=True)
    def _():
        nonlocal n_times
        n_times += 1

    await flush()
    assert n_times == 4

    v.set(2)
    await flush()
    assert n_times == 5

    # Isolates properly
    v = Value(1)
    v2 = Value(1)

    @effect()
    @event(v)
    def _():
        nonlocal n_times
        n_times += v2()

    await flush()
    assert n_times == 6

    v2.set(2)
    await flush()
    assert n_times == 6

    # works with @calc()
    v2 = Value(1)

    @calc()
    @event(lambda: v2(), ignore_init=True)
    def r2b():
        return 1

    @effect()
    def _():
        nonlocal n_times
        n_times += r2b()

    await flush()
    assert n_times == 6

    v2.set(2)
    await flush()
    assert n_times == 7


# ------------------------------------------------------------
# @event() works as expected with async
# ------------------------------------------------------------
@pytest.mark.asyncio
async def test_event_async_decorator():
    n_times = 0

    # By default, runs every time that event expression is _not_ None (ignore_none=True)
    @effect()
    @event(lambda: None, lambda: ActionButtonValue(0))
    async def _():
        nonlocal n_times
        n_times += 1

    await flush()
    assert n_times == 0

    # Unless ignore_none=False
    @effect()
    @event(lambda: None, lambda: ActionButtonValue(0), ignore_none=False)
    async def _():
        nonlocal n_times
        n_times += 1

    await flush()
    assert n_times == 1

    # Or if one of the args is not None
    @effect()
    @event(lambda: None, lambda: ActionButtonValue(0), lambda: True)
    async def _():
        nonlocal n_times
        n_times += 1

    await flush()
    assert n_times == 2

    # Is invalidated properly by reactive values
    v = Value(1)

    @effect()
    @event(v)
    async def _():
        nonlocal n_times
        n_times += 1

    await flush()
    assert n_times == 3

    v.set(1)
    await flush()
    assert n_times == 3

    v.set(2)
    await flush()
    assert n_times == 4

    # Doesn't run on init
    v = Value(1)

    @effect()
    @event(v, ignore_init=True)
    async def _():
        nonlocal n_times
        n_times += 1

    await flush()
    assert n_times == 4

    v.set(2)
    await flush()
    assert n_times == 5

    # Isolates properly
    v = Value(1)
    v2 = Value(1)

    @effect()
    @event(v)
    async def _():
        nonlocal n_times
        n_times += v2()

    await flush()
    assert n_times == 6

    v2.set(2)
    await flush()
    assert n_times == 6

    # works with @calc()
    v2 = Value(1)

    @calc()
    async def r_a():
        await asyncio.sleep(0)  # Make sure the async function yields control
        return 1

    @calc()
    @event(lambda: v2(), r_a, ignore_init=True)
    async def r2b():
        await asyncio.sleep(0)  # Make sure the async function yields control
        return 1

    @effect()
    async def _():
        nonlocal n_times
        await asyncio.sleep(0)
        n_times += await r2b()

    await flush()
    assert n_times == 6

    v2.set(2)
    await flush()
    assert n_times == 7


# ------------------------------------------------------------
# @event() handles silent exceptions in event function
# ------------------------------------------------------------
@pytest.mark.asyncio
async def test_event_silent_exception():
    n_times = 0
    x = Value[bool]()

    @effect()
    @event(x)
    def _():
        nonlocal n_times
        n_times += 1

    await flush()
    assert n_times == 0

    x.set(True)
    await flush()
    assert n_times == 1

    x.unset()
    await flush()
    assert n_times == 1

    x.set(True)
    await flush()
    assert n_times == 2


# ------------------------------------------------------------
# @event() handles silent exceptions in event function, async
# ------------------------------------------------------------
@pytest.mark.asyncio
async def test_event_silent_exception_async():
    n_times = 0
    x = Value[bool]()

    async def req_fn() -> int:
        await asyncio.sleep(0)
        x()
        return 1234

    @effect()
    @event(req_fn)
    async def _():
        await asyncio.sleep(0)
        nonlocal n_times
        n_times += 1

    await flush()
    assert n_times == 0

    x.set(True)
    await flush()
    assert n_times == 1

    x.unset()
    await flush()
    assert n_times == 1

    x.set(True)
    await flush()
    assert n_times == 2


# ------------------------------------------------------------
# @event() throws runtime errors if passed wrong type
# ------------------------------------------------------------
@pytest.mark.asyncio
async def test_event_type_check():
    with pytest.raises(TypeError):
        # Should complain about missing argument to @event().
        @event()
        async def _(): ...

    with pytest.raises(TypeError):
        # Should complain that @event() can't take the result of @Effect (which returns
        # None).
        @event(lambda: 1)  # pyright: ignore[reportArgumentType]
        @effect()
        async def _(): ...

    with pytest.raises(TypeError):
        # Should complain that @event must be applied before @Calc.
        @event(lambda: 1)
        @calc()
        async def _(): ...

    with pytest.raises(TypeError):
        # Should complain that @event must be applied before @render.text. At some point
        # in the future, this may be allowed.
        @event(lambda: 1)  # pyright: ignore[reportArgumentType]
        @render.text
        async def _(): ...

    with pytest.raises(TypeError):
        # Should complain that @event must be applied before @output.
        @event(lambda: 1)  # pyright: ignore[reportArgumentType]
        @render.text
        async def _(): ...

    # These are OK
    @event(lambda: 1)
    async def _(): ...

    @effect()
    @event(lambda: 1)
    async def _(): ...

    @calc()
    @event(lambda: 1)
    async def _(): ...

    @render.text
    @event(lambda: 1)
    async def _(): ...


# ------------------------------------------------------------
# @output() throws runtime errors if passed wrong type
# ------------------------------------------------------------
@pytest.mark.asyncio
async def test_output_type_check():
    conn = MockConnection()
    session = App(ui.TagList(), None)._create_session(conn)
    output = session.output

    with pytest.raises(TypeError):
        # Should complain about bare function
        @output  # pyright: ignore[reportArgumentType,reportUntypedFunctionDecorator]
        def _(): ...

    with pytest.raises(TypeError):
        # Should complain about @event
        @output  # pyright: ignore[reportArgumentType,reportUntypedFunctionDecorator]
        @event(lambda: 1)
        def _(): ...

    with pytest.raises(TypeError):
        # Should complain about @event, even with render.text. Although maybe in the
        # future this will be allowed.
        @output  # pyright: ignore[reportArgumentType,reportUntypedFunctionDecorator]
        @event(lambda: 1)  # pyright: ignore[reportArgumentType]
        @render.text
        def _(): ...

    with pytest.raises(TypeError):
        # Should complain about @Calc
        @output  # pyright: ignore[reportArgumentType,reportUntypedFunctionDecorator]
        @calc
        def _(): ...

    with pytest.raises(TypeError):
        # Should complain about @Effet
        @output  # pyright: ignore[reportArgumentType,reportUntypedFunctionDecorator]
        @effect
        def _(): ...

    @render.text
    def _(): ...

    @render.plot
    @event(lambda: 1)
    def _(): ...


# ------------------------------------------------------------
# @effect()'s .suspend()/.resume() works as expected
# ------------------------------------------------------------
@pytest.mark.asyncio
async def test_effect_pausing():
    a = Value(float(1))

    @calc()
    def funcA():
        return a()

    @effect()
    def obsB():
        funcA()

    # Important: suspend() only affects effect at invalidation time

    # Effects are invalidated at creation time, so it will run once regardless
    # of being suspended
    obsB.suspend()
    await flush()
    assert funcA._exec_count == 1
    assert obsB._exec_count == 1

    # When resuming, if nothing changed, don't do anything
    obsB.resume()
    await flush()
    assert funcA._exec_count == 1
    assert obsB._exec_count == 1

    # Make sure suspended effects do not flush, but do invalidate
    obsB_invalidated = False

    def _():
        nonlocal obsB_invalidated
        obsB_invalidated = True

    obsB.on_invalidate(_)
    obsB.suspend()
    a.set(2)
    await flush()
    assert obsB_invalidated
    assert funcA._exec_count == 1
    assert obsB._exec_count == 1

    obsB.resume()
    a.set(2.5)
    obsB.suspend()
    await flush()
    assert funcA._exec_count == 2
    assert obsB._exec_count == 2

    a.set(3)
    await flush()
    assert funcA._exec_count == 2
    assert obsB._exec_count == 2

    # If on_invalidate() is added _after_ obsB is suspended and the a changes,
    # then it shouldn't get run (on_invalidate runs on invalidation, not on flush)
    a.set(4)
    obsB_invalidated2 = False

    def _():
        nonlocal obsB_invalidated2
        obsB_invalidated2 = True

    obsB.on_invalidate(_)
    obsB.resume()
    await flush()
    assert not obsB_invalidated2
    assert funcA._exec_count == 3
    assert obsB._exec_count == 3

    obsB.suspend()
    a.set(5)
    obsB.destroy()
    obsB.resume()
    await flush()
    assert funcA._exec_count == 3
    assert obsB._exec_count == 3


# ------------------------------------------------------------
# @effect()'s .suspend()/.resume() works as expected (with async)
# ------------------------------------------------------------
@pytest.mark.asyncio
async def test_effect_async_pausing():
    a = Value(float(1))

    @calc()
    async def funcA():
        return a()

    @effect()
    async def obsB():
        await funcA()

    # Important: suspend() only affects effect at invalidation time

    # Effects are invalidated at creation time, so it will run once regardless
    # of being suspended
    obsB.suspend()
    await flush()
    assert funcA._exec_count == 1
    assert obsB._exec_count == 1

    # When resuming, if nothing changed, don't do anything
    obsB.resume()
    await flush()
    assert funcA._exec_count == 1
    assert obsB._exec_count == 1

    # Make sure suspended effects do not flush, but do invalidate
    obsB_invalidated = False

    def _():
        nonlocal obsB_invalidated
        obsB_invalidated = True

    obsB.on_invalidate(_)
    obsB.suspend()
    a.set(2)
    await flush()
    assert obsB_invalidated
    assert funcA._exec_count == 1
    assert obsB._exec_count == 1

    obsB.resume()
    a.set(2.5)
    obsB.suspend()
    await flush()
    assert funcA._exec_count == 2
    assert obsB._exec_count == 2

    a.set(3)
    await flush()
    assert funcA._exec_count == 2
    assert obsB._exec_count == 2

    # If on_invalidate() is added _after_ obsB is suspended and a changes,
    # then it shouldn't get run (on_invalidate runs on invalidation, not on flush)
    a.set(4)
    obsB_invalidated2 = False

    def _():
        nonlocal obsB_invalidated2
        obsB_invalidated2 = True

    obsB.on_invalidate(_)
    obsB.resume()
    await flush()
    assert not obsB_invalidated2
    assert funcA._exec_count == 3
    assert obsB._exec_count == 3

    obsB.suspend()
    a.set(5)
    obsB.destroy()
    obsB.resume()
    await flush()
    assert funcA._exec_count == 3
    assert obsB._exec_count == 3


@pytest.mark.asyncio
async def test_observer_async_suspended_resumed_observers_run_at_most_once():
    a = Value(1)

    @effect()
    async def obs():
        print(a())

    # First flush should run obs once
    assert obs._exec_count == 0
    await flush()
    assert obs._exec_count == 1

    # Modify the dependency at each stage of suspend/resume/flush should still
    # only result in one run of obs()
    a.set(2)
    obs.suspend()
    a.set(3)
    obs.resume()
    a.set(4)
    await flush()
    assert obs._exec_count == 2



from __future__ import annotations

import datetime
import typing
from typing import Literal

from conftest import create_doc_example_core_fixture
from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.run import ShinyAppProc

app = create_doc_example_core_fixture("input_date")


def expect_date(
    date: controller.InputDate,
    value: str | Literal["today"] = "today",
    *,
    autoclose: bool = True,
    datesdisabled: typing.Optional[list[str]] = None,
    daysofweekdisabled: typing.Optional[list[int]] = None,
    format: str = "yyyy-mm-dd",
    language: str = "en",
    max_date: typing.Optional[str] = None,
    min_date: typing.Optional[str] = None,
    startview: str = "month",
    weekstart: int = 0,
    width: typing.Optional[str] = None,
) -> None:
    date.expect_value(str(datetime.date.today()) if value == "today" else value)
    autoclose_str = "true" if autoclose else "false"
    date.expect_autoclose(autoclose_str)
    date.expect_datesdisabled(datesdisabled)
    date.expect_daysofweekdisabled(daysofweekdisabled)
    date.expect_format(format)
    date.expect_language(language)
    date.expect_max_date(max_date)
    date.expect_min_date(min_date)
    date.expect_startview(startview)
    date.expect_weekstart(weekstart)
    date.expect_width(width)


def test_input_date_kitchen(page: Page, app: ShinyAppProc) -> None:
    page.goto(app.url)

    date1 = controller.InputDate(page, "date1")
    date1.expect_label("Has default date:")

    expect_date(date1, "2016-02-29")

    date1.set("2016-02-15")
    expect_date(date1, "2016-02-15")

    expect_date(controller.InputDate(page, "date2"))

    expect_date(controller.InputDate(page, "date3"), "02/29/16", format="mm/dd/yy")

    expect_date(controller.InputDate(page, "date3"), "02/29/16", format="mm/dd/yy")

    expect_date(controller.InputDate(page, "date4"), "2016-02-29")

    expect_date(controller.InputDate(page, "date5"), language="ru", weekstart=1)

    expect_date(controller.InputDate(page, "date6"), startview="decade")

    expect_date(controller.InputDate(page, "date7"), daysofweekdisabled=[1, 2])

    expect_date(
        controller.InputDate(page, "date8"),
        "2016-02-29",
        datesdisabled=["2016-03-01", "2016-03-02"],
    )

    expect_date(
        controller.InputDate(page, "date9"),
        "2016-02-03",
        min_date="2016-02-01",
        max_date="2016-02-29",
    )

    expect_date(controller.InputDate(page, "date10"), width="600px")

    expect_date(controller.InputDate(page, "date11"), autoclose=False)



from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.run import ShinyAppProc


def test_page_navbar_fixed_bottom_inverse_bg(
    page: Page, local_app: ShinyAppProc
) -> None:
    page.goto(local_app.url)

    page_navbar_fixed_bottom_inverse_bg = controller.PageNavbar(
        page, "page_fixed_bottom_inverse_bg"
    )
    page_navbar_fixed_bottom_inverse_bg.expect_position("fixed-bottom")
    page_navbar_fixed_bottom_inverse_bg.expect_inverse(value=True)
    page_navbar_fixed_bottom_inverse_bg.expect_bg("dodgerBlue")



# TODO-future: Add filter of X varaible to reduce the data? (Here we would show "Gentoo" has count 0, rather than remove if no data exists)
# TODO-future: Add brushing to zoom into the plot. The counts should represent the data in the zoomed area. (Single click would zoom out)

from pathlib import Path
from typing import List

import pandas as pd
import seaborn as sns
from colors import bg_palette, palette

from shiny import App, Inputs, Outputs, Session, reactive, render, req, ui

sns.set_theme()

www_dir = Path(__file__).parent.resolve() / "www"

df = pd.read_csv(Path(__file__).parent / "penguins.csv", na_values="NA")
numeric_cols: List[str] = df.select_dtypes(include=["float64"]).columns.tolist()
species: List[str] = df["Species"].unique().tolist()
species.sort()

app_ui = ui.page_sidebar(
    ui.sidebar(
        # Artwork by @allison_horst
        ui.tags.img(src="palmerpenguins.png", width="80%", class_="mt-0 mb-2 mx-auto"),
        ui.input_selectize(
            "xvar",
            "X variable",
            numeric_cols,
            selected="Bill Length (mm)",
        ),
        ui.input_selectize(
            "yvar",
            "Y variable",
            numeric_cols,
            selected="Bill Depth (mm)",
        ),
        ui.input_checkbox_group(
            "species", "Filter by species", species, selected=species
        ),
        ui.hr(),
        ui.input_switch("by_species", "Show species", value=True),
        ui.input_switch("show_margins", "Show marginal plots", value=True),
    ),
    ui.output_ui("value_boxes"),
    ui.output_plot("scatter", fill=True),
    ui.help_text(
        "Artwork by ",
        ui.a("@allison_horst", href="https://twitter.com/allison_horst"),
        class_="text-end",
    ),
)


def server(input: Inputs, output: Outputs, session: Session):
    @reactive.calc
    def filtered_df() -> pd.DataFrame:
        """Returns a pandas data frame that includes only the desired rows"""

        # This calculation "req"uires that at least one species is selected
        req(len(input.species()) > 0)

        # Filter the rows so we only include the desired species
        return df[df["Species"].isin(input.species())]

    @render.plot
    def scatter():
        """Generates a plot for Shiny to display to the user"""

        # The plotting function to use depends on whether margins are desired
        plotfunc = sns.jointplot if input.show_margins() else sns.scatterplot

        plotfunc(
            data=filtered_df(),
            x=input.xvar(),
            y=input.yvar(),
            palette=palette,
            hue="Species" if input.by_species() else None,
            hue_order=species,
            legend=False,
        )

    @render.ui
    def value_boxes():
        df = filtered_df()

        def penguin_value_box(title: str, count: int, bgcol: str, showcase_img: str):
            return ui.value_box(
                title,
                count,
                {"class": "pt-1 pb-0"},
                showcase=ui.fill.as_fill_item(
                    ui.tags.img(
                        {"style": "object-fit:contain;"},
                        src=showcase_img,
                    )
                ),
                theme_color=None,
                style=f"background-color: {bgcol};",
                full_screen=True,
            )

        if not input.by_species():
            return penguin_value_box(
                "Penguins",
                len(df.index),
                bg_palette["default"],
                # Artwork by @allison_horst
                showcase_img="penguins.png",
            )

        value_boxes = [
            penguin_value_box(
                name,
                len(df[df["Species"] == name]),
                bg_palette[name],
                # Artwork by @allison_horst
                showcase_img=f"{name}.png",
            )
            for name in species
            # Only include boxes for _selected_ species
            if name in input.species()
        ]

        return ui.layout_column_wrap(*value_boxes, width=1 / len(value_boxes))


app = App(
    app_ui,
    server,
    static_assets=str(www_dir),
)



from __future__ import annotations

import asyncio
from typing import List

from shiny import App, Inputs, Outputs, Session, reactive, render, ui

app_ui = ui.page_fluid(
    ui.markdown(
        """
        # `session.on_flush` and `session.on_flushed` Reprex

        Verify that `on_flush` and `on_flushed` are called in the correct order, and that they can be cancelled, handle, synchronous functions, and handle asynchronous functions.

        It is not safe to put reactivity inside `session.on_flush()` or `session.on_flushed()` callbacks as the reactive graph is not locked. This app breaks that rule to test the behavior as there is only one user. (If there were multiple users, the reactive graph would not be locked could update while waiting for an async callback.)

        The reprex below will click the button twice (click count is `K`), once after 250ms and another after another 250ms. The expected output is:
        * `a-K-EVENT`
        * `bx-K-first-EVENT`
        * `by-K-first-EVENT`
        * `bx-K-second-EVENT`
        * `by-K-second-EVENT`
        * `c-K-EVENT`


        Even though the `flush` and `flushed` events where mixed when being added, all `flush` events should occur before all `flushed` events.

        Without something to continuously trigger the reactive graph, the `K` value will be `1` less than the click count. To combat this, a reactive event will trigger every 250ms to invoke session `flush` / `flushed` callback.

        ## Automated Reprex:
        """
    ),
    ui.input_action_button("btn", "Click me!"),
    ui.tags.br(),
    ui.tags.span("Counter: "),
    ui.output_text_verbatim("btn_txt", placeholder=True),
    ui.tags.span("All events: "),
    ui.output_text_verbatim("all_txt", placeholder=True),
    ui.tags.span("Flush events: "),
    ui.output_text_verbatim("flush_txt", placeholder=True),
    ui.tags.span("Flushed events: "),
    ui.output_text_verbatim("flushed_txt", placeholder=True),
    ui.tags.span("Session end events (refresh App to add events): "),
    ui.output_text_verbatim("session_end_txt", placeholder=True),
    ui.tags.script(
        """
        $(document).on('shiny:connected', function(event) {
            const n = 250
            document.querySelector("#btn").click();
            setTimeout(function() {
                document.querySelector("#btn").click();
            }, n)
            setTimeout(function() {
                document.querySelector("#btn").click();
            }, 2 * n)
        });
        """
    ),
)

session_ended_messages: List[str] = []


def server(input: Inputs, output: Outputs, session: Session):
    def on_ended_sync(txt: str):
        def _():
            session_ended_messages.append(txt)

        return _

    def on_ended_async(txt: str):
        async def _():
            await asyncio.sleep(0)
            session_ended_messages.append(txt)

        return _

    cancel_on_ended_sync = session.on_ended(
        on_ended_sync("session ended - sync - cancel")
    )
    cancel_on_ended_async = session.on_ended(
        on_ended_async("session ended - async - cancel")
    )
    session.on_ended(on_ended_sync("session ended - sync - test1"))
    session.on_ended(on_ended_async("session ended - async - test2"))
    session.on_ended(on_ended_async("session ended - async - test3"))
    session.on_ended(on_ended_sync("session ended - sync - test4"))

    all_vals: reactive.Value[tuple[str, ...]] = reactive.Value(())
    flush_vals: reactive.Value[tuple[str, ...]] = reactive.Value(())
    flushed_vals: reactive.Value[tuple[str, ...]] = reactive.Value(())

    def call_a(
        vals: reactive.Value[tuple[str, ...]],
        suffix: str,
    ):
        def _():
            with reactive.isolate():
                all_vals.set(all_vals.get() + (f"a-{suffix}",))
                vals.set(vals.get() + (f"a-{suffix}",))

        return _

    def call_b(
        vals: reactive.Value[tuple[str, ...]],
        suffix: str,
    ):
        async def _():
            with reactive.isolate():
                all_vals.set(all_vals.get() + (f"bx-{suffix}",))
                vals.set(vals.get() + (f"bx-{suffix}",))
            await asyncio.sleep(0)
            with reactive.isolate():
                all_vals.set(all_vals.get() + (f"by-{suffix}",))
                vals.set(vals.get() + (f"by-{suffix}",))

        return _

    def call_c(
        vals: reactive.Value[tuple[str, ...]],
        suffix: str,
    ):
        def _():
            with reactive.isolate():
                all_vals.set(all_vals.get() + (f"c-{suffix}",))
                vals.set(vals.get() + (f"c-{suffix}",))

        return _

    # Continuously trigger the reactive graph to ensure that the flush / flushed
    # callbacks are called. If this Effect is not called, then the click counter will
    # always be one higher than the flush/flushed values displayed.
    @reactive.effect
    def _():
        reactive.invalidate_later(0.25)

    @reactive.effect
    @reactive.event(input.btn)
    def _():
        btn_count = input.btn()

        def reset():
            all_vals.set(())
            flush_vals.set(())
            flushed_vals.set(())

        session.on_flush(reset, once=True)

        session.on_flushed(call_a(flushed_vals, f"{btn_count}-flushed"), once=True)
        session.on_flushed(
            call_b(flushed_vals, f"{btn_count}-first-flushed"), once=True
        )
        session.on_flush(call_a(flush_vals, f"{btn_count}-flush"), once=True)
        cancel_b_flush = session.on_flush(
            call_b(flush_vals, f"{btn_count}-cancel-flush"), once=True
        )
        session.on_flush(call_b(flush_vals, f"{btn_count}-first-flush"), once=True)
        session.on_flushed(
            call_b(flushed_vals, f"{btn_count}-second-flushed"), once=True
        )
        session.on_flush(call_b(flush_vals, f"{btn_count}-second-flush"), once=True)
        session.on_flushed(call_c(flushed_vals, f"{btn_count}-flushed"), once=True)
        cancel_c_flushed = session.on_flushed(
            call_c(flushed_vals, f"{btn_count}-cancel-flushed"), once=True
        )
        session.on_flush(call_c(flush_vals, f"{btn_count}-flush"), once=True)

        cancel_b_flush()
        cancel_c_flushed()
        cancel_on_ended_sync()
        cancel_on_ended_async()

    @render.text
    def btn_txt():
        return str(input.btn())

    @render.text
    def all_txt():
        return str(all_vals.get())

    @render.text
    def flush_txt():
        return str(flush_vals.get())

    @render.text
    def flushed_txt():
        return str(flushed_vals.get())

    @render.text
    def session_end_txt():
        return str(session_ended_messages)


app = App(app_ui, server)



from shiny import App, Inputs, Outputs, Session, render, ui

app_ui = ui.page_fluid(ui.output_ui("dynamic_ui"))


def server(input: Inputs, output: Outputs, session: Session):
    @render.ui
    def dynamic_ui():
        @render.text
        def txt():
            return "Hello"

        return txt


app = App(app_ui, server)



from __future__ import annotations

import platform
from typing import Literal, Protocol

from playwright.sync_api import Locator, Page
from playwright.sync_api import expect as playwright_expect

from shiny.render._data_frame import ColumnFilter, ColumnSort

from .._types import AttrValue, ListPatternOrStr, PatternOrStr, StyleValue, Timeout
from ..expect import expect_not_to_have_class, expect_to_have_class
from ..expect._internal import (
    expect_attribute_to_have_value as _expect_attribute_to_have_value,
)
from ..expect._internal import expect_class_to_have_value as _expect_class_to_have_value
from ..expect._internal import expect_style_to_have_value as _expect_style_to_have_value
from ._base import InitLocator, OutputBaseP, UiWithContainer


class _OutputBase:
    """
    Base class for output controls.
    """

    id: str
    """
    The ID of the output control.
    """
    loc: Locator
    """
    Playwright `Locator` of the output control.
    """
    page: Page
    """
    Playwright `Page` of the Shiny app.
    """

    def __init__(
        self,
        page: Page,
        *,
        id: str,
        loc: InitLocator,
    ) -> None:
        self.page = page
        self.id = id

        if isinstance(loc, str):
            loc = page.locator(loc)
        self.loc = loc

    @property
    # TODO; Return type
    def expect(self):
        return playwright_expect(self.loc)


class _OutputTextValue(_OutputBase):
    # cls = "shiny-text-output" + (" noplaceholder" if not placeholder else "")
    # return tags.pre(id=resolve_id(id), class_=cls)

    def expect_value(
        self,
        value: PatternOrStr,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Asserts that the output has the expected value.

        Parameters
        ----------
        value
            The expected value.
        timeout
            The maximum time to wait for the value to appear. Defaults to `None`.
        """
        """Note this function will trim value and output text value before comparing them"""
        self.expect.to_have_text(value, timeout=timeout)


class _OutputContainerP(OutputBaseP, Protocol):
    def expect_container_tag(
        self: OutputBaseP,
        value: Literal["span", "div"] | str,
        *,
        timeout: Timeout = None,
    ) -> None: ...


class _OutputContainerM:
    def expect_container_tag(
        self: OutputBaseP,
        value: Literal["span", "div"] | str,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Asserts that the output has the expected container tag.

        Parameters
        ----------
        value
            The expected container tag.
        timeout
            The maximum time to wait for the container tag to appear. Defaults to `None`.
        """
        loc = self.loc.locator(f"xpath=self::{value}")
        playwright_expect(loc).to_have_count(1, timeout=timeout)


class _OutputInlineContainerM(_OutputContainerM):
    def expect_inline(
        self: _OutputContainerP,
        value: bool = False,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Asserts that the output is inline.

        Parameters
        ----------
        value
            Whether the output is inline.
        timeout
            The maximum time to wait for the output to appear. Defaults to `None`.
        """
        tag_name = "span" if value else "div"
        self.expect_container_tag(tag_name, timeout=timeout)


class OutputText(
    _OutputInlineContainerM,
    _OutputTextValue,
):
    """Controller for :func:`shiny.ui.output_text`."""

    loc: Locator
    """
    Playwright `Locator` of the text output.
    """

    def __init__(
        self,
        page: Page,
        id: str,
    ) -> None:
        """
        Initializes a text output.

        Parameters
        ----------
        page
            Playwright `Page` of the Shiny app.
        id
            The ID of the text output.
        """
        super().__init__(page, id=id, loc=f"#{id}.shiny-text-output")

    def get_value(self, *, timeout: Timeout = None) -> str:
        """
        Gets the text value of the output.

        Parameters
        ----------
        timeout
            The maximum time to wait for the value to appear. Defaults to `None`.
        """
        return self.loc.inner_text(timeout=timeout)


class OutputCode(_OutputTextValue):
    """Controller for :func:`shiny.ui.output_code`."""

    loc: Locator
    """
    Playwright `Locator` of the code output.
    """

    def __init__(self, page: Page, id: str) -> None:
        """
        Initializes a code output.

        Parameters
        ----------
        page
            Playwright `Page` of the Shiny app.
        id
            The ID of the code output.
        """
        super().__init__(page, id=id, loc=f"pre#{id}.shiny-text-output")

    def expect_has_placeholder(
        self,
        value: bool = False,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Asserts that the code output has the expected placeholder.

        Parameters
        ----------
        value
            Whether the code output has a placeholder.
        timeout
            The maximum time to wait for the placeholder to appear. Defaults to `None`.
        """
        _expect_class_to_have_value(
            self.loc,
            class_="noplaceholder",
            has_class=not value,
            timeout=timeout,
        )


class OutputTextVerbatim(_OutputTextValue):
    """Controller for :func:`shiny.ui.output_text_verbatim`."""

    loc: Locator
    """
    Playwright `Locator` of the verbatim text output.
    """

    def __init__(self, page: Page, id: str) -> None:
        """
        Initializes a verbatim text output.

        Parameters
        ----------
        page
            Playwright `Page` of the Shiny app.
        id
            The ID of the verbatim text output.
        """
        super().__init__(page, id=id, loc=f"pre#{id}.shiny-text-output")

    def expect_has_placeholder(
        self,
        value: bool = False,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Asserts that the verbatim text output has the expected placeholder.

        Parameters
        ----------
        value
            Whether the verbatim text output has a placeholder.
        timeout
            The maximum time to wait for the placeholder to appear. Defaults to `None`.
        """
        _expect_class_to_have_value(
            self.loc,
            class_="noplaceholder",
            has_class=not value,
            timeout=timeout,
        )


class _OutputImageBase(_OutputInlineContainerM, _OutputBase):

    loc_img: Locator
    """
    Playwright `Locator` of the image.
    """

    def __init__(
        self,
        page: Page,
        id: str,
        loc_classes: str = "",
    ) -> None:
        """
        Initializes an image output.

        Parameters
        ----------
        page
            Playwright `Page` of the Shiny app.
        id
            The ID of the image.
        loc_classes
            Additional classes to locate the image. Defaults to "".
        """
        super().__init__(
            page,
            id=id,
            loc=f"#{id}.shiny-image-output{loc_classes}",
        )
        self.loc_img = self.loc.locator("img")

    def expect_height(
        self,
        value: StyleValue,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Asserts that the image has the expected height.

        Parameters
        ----------
        value
            The expected height.
        timeout
            The maximum time to wait for the height to appear. Defaults to `None`.
        """
        _expect_style_to_have_value(self.loc, "height", value, timeout=timeout)

    def expect_width(
        self,
        value: StyleValue,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Asserts that the image has the expected width.

        Parameters
        ----------
        value
            The expected width.
        timeout
            The maximum time to wait for the width to appear. Defaults to `None`.
        """
        _expect_style_to_have_value(self.loc, "width", value, timeout=timeout)

    def expect_img_src(
        self,
        value: AttrValue,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Asserts that the image has the expected src.

        Parameters
        ----------
        value
            The expected src.
        timeout
            The maximum time to wait for the src to appear. Defaults to `None`.
        """
        _expect_attribute_to_have_value(self.loc_img, "src", value, timeout=timeout)

    def expect_img_width(
        self,
        value: AttrValue,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Asserts that the image has the expected width.

        Parameters
        ----------
        value
            The expected width.
        timeout
            The maximum time to wait for the width to appear. Defaults to `None`.
        """
        _expect_attribute_to_have_value(self.loc_img, "width", value, timeout=timeout)

    def expect_img_height(
        self,
        value: AttrValue,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Asserts that the image has the expected height.

        Parameters
        ----------
        value
            The expected height.
        timeout
            The maximum time to wait for the height to appear. Defaults to `None`.
        """
        _expect_attribute_to_have_value(self.loc_img, "height", value, timeout=timeout)

    def expect_img_alt(
        self,
        value: AttrValue,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Asserts that the image has the expected alt text.

        Parameters
        ----------
        value
            The expected alt text.
        timeout
            The maximum time to wait for the alt text to appear. Defaults to `None`.
        """
        _expect_attribute_to_have_value(self.loc_img, "alt", value, timeout=timeout)

    # def expect_img_style(
    #     self,
    #     value: AttrValue,
    #     *,
    #     timeout: Timeout = None,
    # ) -> None:
    #     expect_attr(self.loc_img, "style", value, timeout=timeout)


class OutputImage(_OutputImageBase):
    """Controller for :func:`shiny.ui.output_image`."""

    def __init__(self, page: Page, id: str) -> None:
        """
        Initializes an image output.

        Parameters
        ----------
        page
            Playwright `Page` of the Shiny app.
        id
            The ID of the image.
        """
        super().__init__(page, id=id)


class OutputPlot(_OutputImageBase):
    """Controller for :func:`shiny.ui.output_plot`."""

    loc: Locator
    """
    Playwright `Locator` of the plot output.
    """

    def __init__(self, page: Page, id: str) -> None:
        """
        Initializes a plot output.

        Parameters
        ----------
        page
            Playwright `Page` of the Shiny app.
        id
            The ID of the plot.
        """
        super().__init__(page, id=id, loc_classes=".shiny-plot-output")


class OutputUi(_OutputInlineContainerM, _OutputBase):
    """Controller for :func:`shiny.ui.output_ui`."""

    def __init__(self, page: Page, id: str) -> None:
        """
        Initializes a UI output.

        Parameters
        ----------
        page
            Playwright `Page` of the Shiny app.
        id
            The ID of the UI output.
        """
        super().__init__(page, id=id, loc=f"#{id}")

    # TODO-future; Should we try verify that `recalculating` class is not present? Do this for all outputs!
    def expect_empty(self, value: bool, *, timeout: Timeout = None) -> None:
        """
        Asserts that the output is empty.

        Parameters
        ----------
        value
            Whether the output is empty.
        timeout
            The maximum time to wait for the output to be empty. Defaults to `None`.
        """
        if value:
            self.expect.to_be_empty(timeout=timeout)
        else:
            self.expect.not_to_be_empty(timeout=timeout)


class OutputTable(_OutputBase):
    """Controller for :func:`shiny.ui.output_table`."""

    def __init__(self, page: Page, id: str) -> None:
        """
        Initializes a table output.

        Parameters
        ----------
        page
            Playwright `Page` of the Shiny app.
        id
            The ID of the table.
        """
        super().__init__(page, id=id, loc=f"#{id}")

    def expect_cell(
        self,
        value: PatternOrStr,
        row: int,
        col: int,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Asserts that the table cell has the expected text.

        Parameters
        ----------
        value
            The expected text in the cell.
        row
            The row number.
        col
            The column number.
        timeout
            The maximum time to wait for the text to appear. Defaults to `None`.
        """
        if not isinstance(row, int):
            raise TypeError("`row` must be an integer")
        if not isinstance(col, int):
            raise TypeError("`col` must be an integer")
        playwright_expect(
            self.loc.locator(
                f"xpath=./table/tbody/tr[{row}]/td[{col}] | ./table/tbody/tr[{row}]/th[{col}]"
            )
        ).to_have_text(value, timeout=timeout)

    def expect_column_labels(
        self,
        value: ListPatternOrStr | None,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Asserts that the table has the expected column labels.

        Parameters
        ----------
        value
            The expected column labels. If None, it asserts that the table has no column labels.
        timeout
            The maximum time to wait for the column labels to appear. Defaults to `None`.
        """
        if isinstance(value, list) and len(value) == 0:
            value = None

        if value is None:
            playwright_expect(
                self.loc.locator("xpath=./table/thead/tr/th")
            ).to_have_count(0, timeout=timeout)
        else:
            playwright_expect(
                self.loc.locator("xpath=./table/thead/tr/th")
            ).to_have_text(value, timeout=timeout)

    def expect_column_text(
        self,
        col: int,
        # Can't use `None` as we don't know how many rows exist
        value: ListPatternOrStr,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Asserts that the column has the expected text.

        Parameters
        ----------
        col
            The column number.
        value
            The expected text in the column.
        timeout
            The maximum time to wait for the text to appear. Defaults to `None`.
        """
        if not isinstance(col, int):
            raise TypeError("`col` must be an integer")
        playwright_expect(
            self.loc.locator(f"xpath=./table/tbody/tr/td[{col}]")
        ).to_have_text(
            value,
            timeout=timeout,
        )

    def expect_ncol(
        self,
        value: int,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Asserts that the table has the expected number of columns.

        Parameters
        ----------
        value
            The expected number of columns in the table.
        timeout
            The maximum time to wait for the table to have the expected number of columns. Defaults to `None`.
        """
        playwright_expect(
            # self.loc.locator("xpath=./table/thead/tr[1]/(td|th)")
            self.loc.locator("xpath=./table/thead/tr[1]/td | ./table/thead/tr[1]/th")
        ).to_have_count(
            value,
            timeout=timeout,
        )

    def expect_nrow(
        self,
        value: int,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Asserts that the table has the expected number of rows.

        Parameters
        ----------
        value
            The expected number of rows in the table.
        timeout
            The maximum time to wait for the table to have the expected number of rows. Defaults to `None`.
        """
        playwright_expect(self.loc.locator("xpath=./table/tbody/tr")).to_have_count(
            value,
            timeout=timeout,
        )


class OutputDataFrame(UiWithContainer):
    """
    Controller for :func:`shiny.ui.output_data_frame`.
    """

    loc_container: Locator
    """
    Playwright `Locator` for the data frame container.
    """
    loc: Locator
    """
    Playwright `Locator` for the data frame.
    """
    loc_head: Locator
    """
    Playwright `Locator` for the head of the data frame table.
    """
    loc_body: Locator
    """
    Playwright `Locator` for the body of the data frame table.
    """

    def __init__(self, page: Page, id: str) -> None:
        """
        Initializes a new instance of the `OutputDataFrame` class.

        Parameters
        ----------
        page
            Playwright `Page` of the Shiny app.
        id
            The ID of the data frame.
        """
        super().__init__(
            page,
            id=id,
            loc_container=f"#{id}.html-fill-item",
            loc="> div > div.shiny-data-grid",
        )
        self.loc_table = self.loc.locator("> table")
        self.loc_head = self.loc_table.locator("> thead")
        self.loc_body = self.loc_table.locator("> tbody")
        self.loc_column_filter = self.loc_head.locator(
            "> tr.filters > th:not(.table-corner)"
        )
        self.loc_column_label = self.loc_head.locator(
            "> tr:not(.filters) > th:not(.table-corner)"
        )

    def cell_locator(self, row: int, col: int) -> Locator:
        """
        Returns the locator for a specific cell in the data frame.

        Parameters
        ----------
        row
            The row number of the cell.
        col
            The column number of the cell.
        """

        return (
            # Find the direct row
            self.loc_body.locator(f"> tr[data-index='{row}']")
            # Find all direct td's and th's (these are independent sets)
            .locator("> td, > th")
            # Remove all results that contain the `row-number` class
            .locator(
                # self
                "xpath=.",
                has=self.page.locator(
                    "xpath=self::*[not(contains(@class, 'row-number'))]"
                ),
            )
            # Return the first result
            .nth(col)
        )

    # TODO-barret; Should this be called `expect_row_count()`?
    def expect_nrow(self, value: int, *, timeout: Timeout = None):
        """
        Expects the number of rows in the data frame.

        Parameters
        ----------
        value
            The expected number of rows.
        timeout
            The maximum time to wait for the expectation to pass. Defaults to `None`.
        """
        playwright_expect(self.loc_table).to_have_attribute(
            "aria-rowcount",
            str(value),
            timeout=timeout,
        )

    def expect_selected_num_rows(self, value: int, *, timeout: Timeout = None):
        """
        Expects the number of selected rows in the data frame.

        Parameters
        ----------
        value
            The expected number of selected rows.
        timeout
            The maximum time to wait for the expectation to pass. Defaults to `None`.
        """
        playwright_expect(
            self.loc_body.locator("tr[aria-selected=true]")
        ).to_have_count(value, timeout=timeout)

    def expect_selected_rows(self, rows: list[int], *, timeout: Timeout = None):
        """
        Expects the specified rows to be selected.

        Parameters
        ----------
        rows
            The row numbers.
        timeout
            The maximum time to wait for the expectation to pass. Defaults to `None`.
        """
        # * given container...
        # * Add that container has all known rows
        # * Verify that selected row count is of size N
        big_loc = self.loc_body
        assert len(rows) > 0
        for row in rows:
            big_loc = big_loc.locator(
                "xpath=.",  # return "self"
                has=self.page.locator(
                    f"> tr[data-index='{row}'][aria-selected='true']"
                ),
            )

        try:
            playwright_expect(
                big_loc.locator("> tr[aria-selected='true']")
            ).to_have_count(len(rows), timeout=timeout)
        except AssertionError as e:
            # Debug expections

            # Expecting container to exist (count = 1)
            playwright_expect(self.loc_body).to_have_count(1, timeout=timeout)

            for row in rows:
                # Expecting item `{item}` to exist in container
                # Perform exact matches on strings.
                playwright_expect(
                    # Simple approach as position is not needed
                    self.loc_body.locator(
                        f"> tr[aria-selected='true'][data-index='{row}']",
                    )
                ).to_have_count(1, timeout=timeout)

            # Could not find the reason why. Raising the original error.
            raise e

    def _expect_row_focus_state(
        self, in_focus: bool = True, *, row: int, timeout: Timeout = None
    ):
        """
        Expects the focus state of the specified row.

        Parameters
        ----------
        row
            The row number.
        in_focus
            `True` if the row is expected to be in focus, `False` otherwise.
        timeout
            The maximum time to wait for the expectation to pass. Defaults to `None`.
        """
        if in_focus:
            playwright_expect(
                self.loc_body.locator(f"> tr[data-index='{row}']")
            ).to_be_focused(timeout=timeout)
        else:
            playwright_expect(
                self.loc_body.locator(f"> tr[data-index='{row}']")
            ).not_to_be_focused(timeout=timeout)

    def expect_cell(
        self,
        value: PatternOrStr,
        *,
        row: int,
        col: int,
        timeout: Timeout = None,
    ) -> None:
        """
        Expects the cell in the data frame to have the specified text.

        Parameters
        ----------
        value
            The expected text in the cell.
        row
            The row number of the cell.
        col
            The column number of the cell.
        timeout
            The maximum time to wait for the expectation to pass. Defaults to `None`.
        """
        if not isinstance(row, int):
            raise TypeError("`row` must be an integer.")
        if not isinstance(col, int):
            raise TypeError("`col` must be an integer.")
        self._cell_scroll_if_needed(row=row, col=col, timeout=timeout)
        playwright_expect(self.cell_locator(row, col)).to_have_text(
            value, timeout=timeout
        )

    def expect_column_labels(
        self,
        value: ListPatternOrStr | None,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Expects the column labels in the data frame.

        Parameters
        ----------
        value
            The expected column labels.

            Note: None if the column labels are expected to not exist.
        timeout
            The maximum time to wait for the expectation to pass. Defaults to `None`.
        """
        if isinstance(value, list) and len(value) == 0:
            value = None

        if value is None:
            playwright_expect(self.loc_column_label).to_have_count(0, timeout=timeout)
        else:
            playwright_expect(self.loc_column_label).to_have_text(
                value, timeout=timeout
            )

    def _cell_scroll_if_needed(self, *, row: int, col: int, timeout: Timeout):
        """
        Scrolls the cell into view if needed.

        Parameters
        ----------
        row
            The row number of the cell.
        col
            The column number of the cell.
        timeout
            The maximum time to wait for the action to complete.
        """
        # Check first and last row data-index and make sure `row` is included

        cell = self.cell_locator(row=row, col=col)

        # Scroll down if top number is larger
        while not cell.is_visible(timeout=timeout):
            first_row = self.loc_body.locator("> tr[data-index]").first
            first_row_index = first_row.get_attribute("data-index")
            if first_row_index is None:
                break
            if int(first_row_index) >= row:
                first_row.scroll_into_view_if_needed(timeout=timeout)
            else:
                # First row index is lower than `row`
                break
        # Scroll up if bottom number is smaller
        while not cell.is_visible(timeout=timeout):
            last_row = self.loc_body.locator("> tr[data-index]").last
            last_row_index = last_row.get_attribute("data-index")
            if last_row_index is None:
                break
            if int(last_row_index) <= row:
                last_row.scroll_into_view_if_needed(timeout=timeout)
            else:
                # Last row index is higher than `row`
                break
        cell.scroll_into_view_if_needed(timeout=timeout)

    def _expect_column_label(
        self,
        value: ListPatternOrStr,
        *,
        col: int,
        timeout: Timeout = None,
    ) -> None:
        """
        Expects the text in the specified column of the data frame.

        Parameters
        ----------
        value
            The expected text in the column.
        col
            The column number.
        timeout
            The maximum time to wait for the expectation to pass. Defaults to `None`.
        """
        if not isinstance(col, int):
            raise TypeError("`col` must be an integer.")
        # It's zero based, nth(0) selects the first element.
        playwright_expect(self.loc_column_label.nth(col - 1)).to_have_text(
            value,
            timeout=timeout,
        )

    def expect_ncol(
        self,
        value: int,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Expects the number of columns in the data frame.

        Parameters
        ----------
        value
            The expected number of columns.
        timeout
            The maximum time to wait for the expectation to pass. Defaults to `None`.
        """
        playwright_expect(self.loc_column_label).to_have_count(
            value,
            timeout=timeout,
        )

    def expect_cell_class(
        self,
        value: str,
        *,
        row: int,
        col: int,
        timeout: Timeout = None,
    ) -> None:
        """
        Expects the class of the cell

        Parameters
        ----------
        value
            The expected class of the cell.
        row
            The row number of the cell.
        col
            The column number of the cell.
        timeout
            The maximum time to wait for the expectation to pass. Defaults to `None`.
        """
        expect_to_have_class(
            self.cell_locator(row=row, col=col),
            value,
            timeout=timeout,
        )

    def select_rows(
        self,
        value: list[int],
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Selects the rows in the data frame.

        Parameters
        ----------
        value
            The list of row numbers to select.
        timeout
            The maximum time to wait for the action to complete. Defaults to `None`.
        """
        if len(value) > 1:
            value = sorted(value)
            # check if the items in the row contain all numbers from index 0 to index -1
            if value == list(range(value[0], value[-1] + 1)):
                self.page.keyboard.down("Shift")
                self.cell_locator(row=value[0], col=0).click(timeout=timeout)
                self.cell_locator(row=value[-1], col=0).click(timeout=timeout)
                self.page.keyboard.up("Shift")
            else:
                # if operating system is MacOs use Meta (Cmd) else use Ctrl key
                if platform.system() == "Darwin":
                    self.page.keyboard.down("Meta")
                else:
                    self.page.keyboard.down("Control")
                for row in value:
                    self._cell_scroll_if_needed(row=row, col=0, timeout=timeout)
                    self.cell_locator(row=row, col=0).click(timeout=timeout)
                if platform.system() == "Darwin":
                    self.page.keyboard.up("Meta")
                else:
                    self.page.keyboard.up("Control")
        else:
            self.cell_locator(row=value[0], col=0).click(timeout=timeout)

    def expect_class_state(
        self,
        value: str,
        *,
        row: int,
        col: int,
        timeout: Timeout = None,
    ):
        """
        Expects the state of the class in the data frame.

        Parameters
        ----------
        value
            The expected state of the class.
        row
            The row number of the cell.
        col
            The column number of the cell.
        timeout
            The maximum time to wait for the expectation to pass. Defaults to `None`.
        """
        if value == "ready":
            playwright_expect(self.cell_locator(row=row, col=col)).not_to_have_class(
                "cell-edit-editing", timeout=timeout
            )
        elif value == "editing":
            self.expect_cell_class("cell-edit-editing", row=row, col=col)
        elif value == "saving":
            self.expect_cell_class("cell-edit-saving", row=row, col=col)
        elif value == "failure":
            self.expect_cell_class("cell-edit-failure", row=row, col=col)
        elif value == "success":
            self.expect_cell_class("cell-edit-success", row=row, col=col)
        else:
            raise ValueError(
                "Invalid state. Select one of 'success', 'failure', 'saving', 'editing', 'ready'"
            )

    def _edit_cell_no_save(
        self,
        text: str,
        *,
        row: int,
        col: int,
        timeout: Timeout = None,
    ) -> None:
        """
        Edits the cell in the data frame.

        Parameters
        ----------
        value
            The value to edit in the cell.
        row
            The row number of the cell.
        col
            The column number of the cell.
        timeout
            The maximum time to wait for the action to complete. Defaults to `None`.
        """
        cell = self.cell_locator(row=row, col=col)

        self._cell_scroll_if_needed(row=row, col=col, timeout=timeout)
        cell.dblclick(timeout=timeout)
        cell.locator("> textarea").fill(text)

    def set_sort(
        self,
        sort: int | ColumnSort | list[int | ColumnSort] | None,
        *,
        timeout: Timeout = None,
    ):
        """
        Set or modify the sorting of columns in a table or grid component.
        This method allows setting single or multiple column sorts, or resetting the sort order.

        Parameters
        ----------
        sort
            The sorting configuration to apply. Can be one of the following:
                * `int`: Index of the column to sort by (ascending order by default).
                * `ColumnSort`: A dictionary specifying a single column sort with 'col' and 'desc' keys.
                * `list[int | ColumnSort]`: A list of ints or dictionaries for multi-column sorting.
                * `None`: No sorting applied (not implemented in the current code).

            If a `desc` values is provided within your `ColumnSort` shaped dictionary, then the direction will be set to that value. If no `desc` value is provided, the column will be sorted in the default sorting order.
        timeout
            The maximum time to wait for the action to complete. Defaults to `None`.
        """

        def click_loc(loc: Locator, *, shift: bool = False):
            clickModifier: list[Literal["Shift"]] | None = (
                ["Shift"] if bool(shift) else None
            )
            loc.click(
                timeout=timeout,
                modifiers=clickModifier,
            )
            # Wait for arrows to react a little bit
            # This could possible be changed to a `wait_for_change`, but 150ms should be fine
            self.page.wait_for_timeout(150)

        # Reset arrow sorting by clicking on the arrows until none are found
        sortingArrows = self.loc_column_label.locator("svg.sort-arrow")
        while sortingArrows.count() > 0:
            click_loc(sortingArrows.first)

        # Quit early if no sorting is needed
        if sort is None:
            return

        if isinstance(sort, int) | isinstance(sort, dict) and not isinstance(
            sort, list
        ):
            sort = [sort]

        if not isinstance(sort, list):
            raise ValueError(
                "Invalid sort value. Must be an int, ColumnSort, list[ColumnSort], or None."
            )

        # For every sorting info...
        for sort_info, i in zip(sort, range(len(sort))):
            # TODO-barret-future; assert column does not have `cell-html` class
            shift = i > 0

            if isinstance(sort_info, int):
                sort_info = {"col": sort_info}

            # Verify ColumnSortInfo
            assert isinstance(
                sort_info, dict
            ), f"Invalid sort value at position {i}. Must be an int, ColumnSort, list[ColumnSort], or None."
            assert (
                "col" in sort_info
            ), f"Column index (`col`) at position {i} is required for sorting."

            sort_col = self.loc_column_label.nth(sort_info["col"])
            expect_not_to_have_class(sort_col, "header-html")

            # If no `desc` key is found, click the column to sort and move on
            if "desc" not in sort_info:
                click_loc(sort_col, shift=shift)
                continue

            # "desc" in sort_info
            desc_val = bool(sort_info["desc"])
            sort_col.scroll_into_view_if_needed()
            for _ in range(2):
                if desc_val:
                    # If a descending is found, stop clicking
                    if sort_col.locator("svg.sort-arrow-down").count() > 0:
                        break
                else:
                    # If a ascending is found, stop clicking
                    if sort_col.locator("svg.sort-arrow-up").count() > 0:
                        break
                click_loc(sort_col, shift=shift)

    # TODO-karan-test: Add support for a list of columns ? If so, all other columns should be reset
    def set_filter(
        self,
        # TODO-barret support array of filters
        filter: ColumnFilter | list[ColumnFilter] | None,
        *,
        timeout: Timeout = None,
    ):
        """
        Set or reset filters for columns in a table or grid component.
        This method allows setting string filters, numeric range filters, or clearing all filters.

        Parameters
        ----------
        filter
            The filter to apply. Can be one of the following:
                * `None`: Resets all filters.
                * `ColumnFilterStr`: A dictionary specifying a string filter with 'col' and 'value' keys.
                * `ColumnFilterNumber`: A dictionary specifying a numeric range filter with 'col' and 'value' keys.
        timeout
            The maximum time to wait for the action to complete. Defaults to `None`.
        """
        # reset all filters
        all_input_locs = self.loc_column_filter.locator("> input, > div > input")
        for i in range(all_input_locs.count()):
            input_el = all_input_locs.nth(i)
            input_el.fill("", timeout=timeout)

        if filter is None:
            return

        if isinstance(filter, dict):
            filter = [filter]

        if not isinstance(filter, list):
            raise ValueError(
                "Invalid filter value. Must be a ColumnFilter, list[ColumnFilter], or None."
            )

        for filterInfo in filter:
            if "col" not in filterInfo:
                raise ValueError("Column index (`col`) is required for filtering.")

            if "value" not in filterInfo:
                raise ValueError("Filter value (`value`) is required for filtering.")

            filterColumn = self.loc_column_filter.nth(filterInfo["col"])

            if isinstance(filterInfo["value"], str):
                filterColumn.locator("> input").fill(filterInfo["value"])
            elif isinstance(filterInfo["value"], (tuple, list)):
                header_inputs = filterColumn.locator("> div > input")
                if filterInfo["value"][0] is not None:
                    header_inputs.nth(0).fill(
                        str(filterInfo["value"][0]),
                        timeout=timeout,
                    )
                if filterInfo["value"][1] is not None:
                    header_inputs.nth(1).fill(
                        str(filterInfo["value"][1]),
                        timeout=timeout,
                    )
            else:
                raise ValueError(
                    "Invalid filter value. Must be a string or a tuple/list of two numbers."
                )

    def set_cell(
        self,
        text: str,
        *,
        row: int,
        col: int,
        finish_key: (
            Literal["Enter", "Shift+Enter", "Tab", "Shift+Tab", "Escape"] | None
        ) = None,
        timeout: Timeout = None,
    ) -> None:
        """
        Saves the value of the cell in the data frame.

        Parameters
        ----------
        text
            The key to save the value of the cell.
        row
            The row number of the cell.
        col
            The column number of the cell.
        finish_key
            The key to save the value of the cell. If `None` (the default), no key will
            be pressed and instead the page body will be clicked.
        timeout
            The maximum time to wait for the action to complete. Defaults to `None`.
        """
        self._edit_cell_no_save(text, row=row, col=col, timeout=timeout)
        if finish_key is None:
            self.page.locator("body").click()
        else:
            self.cell_locator(row=row, col=col).locator("> textarea").press(finish_key)

    def expect_cell_title(
        self,
        value: str,
        *,
        row: int,
        col: int,
        timeout: Timeout = None,
    ) -> None:
        """
        Expects the validation message of the cell in the data frame, which will be in
        the `title` attribute of the element.

        Parameters
        ----------
        value
            The expected validation message of the cell.
        row
            The row number of the cell.
        col
            The column number of the cell.
        timeout
            The maximum time to wait for the expectation to pass. Defaults to `None`.
        """
        playwright_expect(self.cell_locator(row=row, col=col)).to_have_attribute(
            name="title", value=value, timeout=timeout
        )



from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.run import ShinyAppProc


def test_checkbox_group_kitchen(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    default = controller.InputCheckboxGroup(page, "default")
    default.expect_label("Basic Checkbox Group")
    default.expect_selected([])
    controller.OutputCode(page, "default_txt").expect_value("()")

    selected = controller.InputCheckboxGroup(page, "selected")
    selected.expect_selected(["Option B", "Option C"])
    controller.OutputCode(page, "selected_txt").expect_value("('Option B', 'Option C')")

    width = controller.InputCheckboxGroup(page, "width")
    width.expect_width("30px")

    inline = controller.InputCheckboxGroup(page, "inline")
    inline.expect_inline(True)
    inline_txt = controller.OutputCode(page, "inline_txt")
    inline_txt.expect_value("()")
    # Set in wrong order
    inline.set(["Option D", "Option A"])
    inline.expect_selected(["Option A", "Option D"])
    inline_txt.expect_value("('Option A', 'Option D')")

    dict_values = controller.InputCheckboxGroup(page, "dict_values")
    dict_values.expect_selected([])
    dict_values.set(["value1", "value4"])
    dict_values.expect_selected(["value1", "value4"])
    dict_values_txt = controller.OutputCode(page, "dict_values_txt")
    dict_values_txt.expect_value("('value1', 'value4')")



from __future__ import annotations

import copy
import os
import secrets
from contextlib import AsyncExitStack, asynccontextmanager
from inspect import signature
from pathlib import Path
from typing import Any, Callable, Mapping, Optional, TypeVar, cast

import starlette.applications
import starlette.exceptions
import starlette.middleware
import starlette.routing
import starlette.websockets
from htmltools import (
    HTMLDependency,
    HTMLDocument,
    HTMLTextDocument,
    RenderedHTML,
    Tag,
    TagList,
)
from starlette.requests import Request
from starlette.responses import HTMLResponse, JSONResponse, Response
from starlette.types import ASGIApp, Message, Receive, Scope, Send

from ._autoreload import InjectAutoreloadMiddleware, autoreload_url
from ._connection import Connection, StarletteConnection
from ._error import ErrorMiddleware
from ._shinyenv import is_pyodide
from ._utils import guess_mime_type, is_async_callable, sort_keys_length
from .html_dependencies import jquery_deps, require_deps, shiny_deps
from .http_staticfiles import FileResponse, StaticFiles
from .session._session import AppSession, Inputs, Outputs, Session, session_context

T = TypeVar("T")

# Default values for App options.
LIB_PREFIX: str = "lib/"
SANITIZE_ERRORS: bool = False
SANITIZE_ERROR_MSG: str = (
    "An error has occurred. Check your logs or contact the app author for clarification."
)


class App:
    """
    Create a Shiny app instance.

    Parameters
    ----------
    ui
        The UI definition for the app (e.g., a call to :func:`~shiny.ui.page_fluid` or
        similar, with layouts and controls nested inside). You can
        also pass a function that takes a :class:`~starlette.requests.Request` and
        returns a UI definition, if you need the UI definition to be created dynamically
        for each pageview.
    server
        A function which is called once for each session, ensuring that each session is
        independent.
    static_assets
        Static files to be served by the app. If this is a string or Path object, it
        must be a directory, and it will be mounted at `/`. If this is a dictionary,
        each key is a mount point and each value is a file or directory to be served at
        that mount point.
    debug
        Whether to enable debug mode.

    Examples
    --------

    ```{python}
    #| eval: false
    from shiny import  App, Inputs, Outputs, Session, ui

    app_ui = ui.page_fluid("Hello Shiny!")

    def server(input: Inputs, output: Outputs, session: Session):
        pass

    app = App(app_ui, server)
    ```
    """

    lib_prefix: str = "lib/"
    """
    A path prefix to place before all HTML dependencies processed by
    ``register_web_dependency()``.
    """

    sanitize_errors: bool = False
    """
    Whether or not to show a generic message (``SANITIZE_ERRORS=True``) or the actual
    message (``SANITIZE_ERRORS=False``) in the app UI when an error occurs. This flag
    may default to ``True`` in some production environments (e.g., Posit Connect).
    """

    sanitize_error_msg: str = (
        "An error has occurred. Check your logs or contact the app author for clarification."
    )
    """
    The message to show when an error occurs and ``SANITIZE_ERRORS=True``.
    """

    ui: RenderedHTML | Callable[[Request], Tag | TagList]
    server: Callable[[Inputs, Outputs, Session], None]

    def __init__(
        self,
        ui: Tag | TagList | Callable[[Request], Tag | TagList] | Path,
        server: (
            Callable[[Inputs], None] | Callable[[Inputs, Outputs, Session], None] | None
        ),
        *,
        static_assets: Optional[str | Path | Mapping[str, str | Path]] = None,
        debug: bool = False,
    ) -> None:
        # Used to store callbacks to be called when the app is shutting down (according
        # to the ASGI lifespan protocol)
        self._exit_stack = AsyncExitStack()

        if server is None:
            self.server = noop_server_fn
        elif len(signature(server).parameters) == 1:
            self.server = wrap_server_fn_with_output_session(
                cast(Callable[[Inputs], None], server)
            )
        elif len(signature(server).parameters) == 3:
            self.server = cast(Callable[[Inputs, Outputs, Session], None], server)
        else:
            raise ValueError(
                "`server` must have 1 (Inputs) or 3 parameters (Inputs, Outputs, Session)"
            )

        self._debug: bool = debug

        # Settings that the user can change after creating the App object.
        self.lib_prefix: str = LIB_PREFIX
        self.sanitize_errors: bool = SANITIZE_ERRORS
        self.sanitize_error_msg: str = SANITIZE_ERROR_MSG

        if static_assets is None:
            static_assets = {}

        if isinstance(static_assets, Mapping):
            static_assets_map = {k: Path(v) for k, v in static_assets.items()}
        else:
            static_assets_map = {"/": Path(static_assets)}

        for _, static_asset_path in static_assets_map.items():
            if not static_asset_path.is_absolute():
                raise ValueError(
                    f'static_assets must be an absolute path: "{static_asset_path}".'
                    " Consider using one of the following instead:\n"
                    f'  os.path.join(os.path.dirname(__file__), "{static_asset_path}")  OR'
                    f'  pathlib.Path(__file__).parent/"{static_asset_path}"'
                )

        # Sort the static assets keys by descending length, to ensure that the most
        # specific paths are mounted first. Suppose there are mounts "/foo" and "/". If
        # "/" is first in the dict, then requests to "/foo/file.html" will never reach
        # the second mount. We need to put "/foo" first and "/" second so that it will
        # actually look in the "/foo" mount.
        static_assets_map = sort_keys_length(static_assets_map, descending=True)
        self._static_assets: dict[str, Path] = static_assets_map

        self._sessions: dict[str, AppSession] = {}

        self._sessions_needing_flush: dict[int, AppSession] = {}

        self._registered_dependencies: dict[str, HTMLDependency] = {}
        self._dependency_handler = starlette.routing.Router()

        for mount_point, static_asset_path in self._static_assets.items():
            self._dependency_handler.routes.append(
                create_static_asset_route(mount_point, static_asset_path)
            )

        starlette_app = self.init_starlette_app()

        self.starlette_app = starlette_app

        if is_uifunc(ui):
            if is_async_callable(cast(Callable[[Request], Any], ui)):
                raise TypeError("App UI cannot be a coroutine function")
            # Dynamic UI: just store the function for later
            self.ui = cast("Callable[[Request], Tag | TagList]", ui)
        elif isinstance(ui, Path):
            if not ui.is_absolute():
                raise ValueError("Path to UI must be absolute")

            self.ui = self._render_page_from_file(ui, lib_prefix=self.lib_prefix)

        else:
            # Static UI: render the UI now and save the results
            self.ui = self._render_page(
                cast("Tag | TagList", ui), lib_prefix=self.lib_prefix
            )

    def init_starlette_app(self) -> starlette.applications.Starlette:
        routes: list[starlette.routing.BaseRoute] = [
            starlette.routing.WebSocketRoute("/websocket/", self._on_connect_cb),
            starlette.routing.Route("/", self._on_root_request_cb, methods=["GET"]),
            starlette.routing.Route(
                "/session/{session_id}/{action}/{subpath:path}",
                self._on_session_request_cb,
                methods=["GET", "POST"],
            ),
            starlette.routing.Mount("/", app=self._dependency_handler),
        ]
        middleware: list[starlette.middleware.Middleware] = []
        if autoreload_url():
            shared_dir = os.path.join(os.path.dirname(__file__), "www", "shared")
            routes.insert(
                0,
                starlette.routing.Mount(
                    "/__shared",
                    app=StaticFiles(directory=shared_dir),
                ),
            )
            middleware.append(
                starlette.middleware.Middleware(InjectAutoreloadMiddleware)
            )
        # In Pyodide mode, an HTTPException(404) being thrown resulted in
        # some default error handler (that happened not to be async) being
        # run in a threadpool, which Pyodide could not handle. So in Pyodide
        # mode, install our own async error handler at the outermost layer
        # that we can.
        if is_pyodide:
            middleware.append(starlette.middleware.Middleware(ErrorMiddleware))

        starlette_app = starlette.applications.Starlette(
            routes=routes,
            middleware=middleware,
            lifespan=self._lifespan,
        )

        return starlette_app

    @asynccontextmanager
    async def _lifespan(self, app: starlette.applications.Starlette):
        async with self._exit_stack:
            yield

    def _create_session(self, conn: Connection) -> AppSession:
        id = secrets.token_hex(32)
        session = AppSession(self, id, conn, debug=self._debug)
        self._sessions[id] = session
        return session

    def _remove_session(self, session: AppSession | str) -> None:
        if isinstance(session, AppSession):
            session = session.id

        if self._debug:
            print(f"remove_session: {session}", flush=True)
        del self._sessions[session]

    def run(self, **kwargs: object) -> None:
        """
        Run the app.

        Parameters
        ----------
        **kwargs
            Keyword arguments passed to :func:`~shiny.run_app`.
        """
        from ._main import run_app

        run_app(self, **kwargs)  # pyright: ignore[reportArgumentType]

    # ASGI entrypoint. Handles HTTP, WebSocket, and lifespan.
    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        await self.starlette_app(scope, receive, send)

    def on_shutdown(self, callback: Callable[[], None]) -> Callable[[], None]:
        """
        Register a callback to be called when the app is shutting down. This can be
        useful for cleaning up app-wide resources, like connection pools, temporary
        directories, worker threads/processes, etc.

        Parameters
        ----------
        callback
            The callback to call. It should take no arguments, and any return value will
            be ignored. Try not to raise an exception in the callback, as exceptions
            during cleanup can hide the original exception that caused the app to shut
            down.

        Returns
        -------
        :
            The callback, to allow this method to be used as a decorator.
        """
        return self._exit_stack.callback(callback)

    async def call_pyodide(self, scope: Scope, receive: Receive, send: Send) -> None:
        """
        Communicate with pyodide.

        Warning
        -------
        This method is not intended for public usage. It's exported for use by
        shinylive.
        """

        # TODO: Pretty sure there are objects that need to be destroy()'d here?
        scope = cast(Any, scope).to_py()

        # ASGI requires some values to be byte strings, not character strings. Those are
        # not that easy to create in JavaScript, so we let the JS side pass us strings
        # and we convert them to bytes here.
        if "headers" in scope:
            # JS doesn't have `bytes` so we pass as strings and convert here
            scope["headers"] = [
                [value.encode("latin-1") for value in header]
                for header in scope["headers"]
            ]
        if "query_string" in scope and scope["query_string"]:
            scope["query_string"] = scope["query_string"].encode("latin-1")
        if "raw_path" in scope and scope["raw_path"]:
            scope["raw_path"] = scope["raw_path"].encode("latin-1")

        async def rcv() -> Message:
            event = await receive()
            return cast(Message, cast(Any, event).to_py())

        async def snd(event: Message):
            await send(event)

        await self(scope, rcv, snd)

    async def stop(self) -> None:
        """
        Stop the app (i.e., close all sessions).

        See Also
        --------
        * :func:`~shiny.Session.close`
        """
        # convert to list to avoid modifying the dict while iterating over it, which
        # throws an error
        for session in list(self._sessions.values()):
            await session.close()

    # ==========================================================================
    # Connection callbacks
    # ==========================================================================
    async def _on_root_request_cb(self, request: Request) -> Response:
        """
        Callback passed to the ConnectionManager which is invoked when a HTTP
        request for / occurs.
        """
        ui: RenderedHTML
        if callable(self.ui):
            ui = self._render_page(self.ui(request), self.lib_prefix)
        else:
            ui = self.ui
        return HTMLResponse(content=ui["html"])

    async def _on_connect_cb(self, ws: starlette.websockets.WebSocket) -> None:
        """
        Callback which is invoked when a new WebSocket connection is established.
        """
        await ws.accept()
        conn = StarletteConnection(ws)
        session = self._create_session(conn)

        await session._run()

    async def _on_session_request_cb(self, request: Request) -> ASGIApp:
        """
        Callback passed to the ConnectionManager which is invoked when a HTTP
        request for /session/* occurs.
        """
        session_id: str = request.path_params["session_id"]  # type: ignore
        action: str = request.path_params["action"]  # type: ignore
        subpath: str = request.path_params["subpath"]  # type: ignore

        if session_id in self._sessions:
            session: AppSession = self._sessions[session_id]
            with session_context(session):
                return await session._handle_request(request, action, subpath)

        return JSONResponse({"detail": "Not Found"}, status_code=404)

    # ==========================================================================
    # Flush
    # ==========================================================================
    def _request_flush(self, session: AppSession) -> None:
        # TODO: Until we have reactive domains, because we can't yet keep track
        # of which sessions need a flush.
        pass
        # self._sessions_needing_flush[session.id] = session

    # ==========================================================================
    # HTML Dependency stuff
    # ==========================================================================
    def _ensure_web_dependencies(self, deps: list[HTMLDependency]) -> None:
        for dep in deps:
            self._register_web_dependency(dep)

    def _register_web_dependency(self, dep: HTMLDependency) -> None:
        # If the dependency has been seen before, quit early.

        # Even if the htmldependency version is higher or lower, the HTML being sent to
        # the user is requesting THIS dependency. Therefore, it should be available to
        # the user independent of any previous versions of the dependency being served.

        # Note: htmltools does de-duplicate dependencies and finds the highest version
        # to return. However, dynamic UI and callable UI do not run through the same
        # filter over time. When using callable UI functions, UI dependencies are reset
        # on refresh. So if a dependency makes it here, it is not necessarily the
        # highest version served over time but is the highest version for this
        # particular UI. Therefore, serve it must be served.
        dep_name = html_dep_name(dep)
        if dep_name in self._registered_dependencies:
            return

        # For HTMLDependencies that have sources on disk, mount the source dir.
        # (Some HTMLDependencies only carry head content, and have no source on disk.)
        if dep.source:
            paths = dep.source_path_map(lib_prefix=self.lib_prefix)
            if paths["source"] != "":
                self._dependency_handler.routes.insert(
                    0,
                    starlette.routing.Mount(
                        "/" + paths["href"],
                        StaticFiles(directory=paths["source"]),
                        name=dep_name,
                    ),
                )

        self._registered_dependencies[dep_name] = dep

    def _render_page(self, ui: Tag | TagList, lib_prefix: str) -> RenderedHTML:
        ui_res = copy.copy(ui)
        # Use presence of the Bootstrap dependency as a signal that the UI uses a
        # shiny.ui.page_*() function, in which case the Shiny CSS is already included.
        has_bootstrap = any(
            [dep.name == "bootstrap" for dep in ui_res.get_dependencies()]
        )
        # Make sure requirejs, jQuery, and Shiny come before any other dependencies.
        # (see require_deps() for a comment about why we even include it)
        ui_res.insert(
            0,
            [require_deps(), jquery_deps(), *shiny_deps(include_css=not has_bootstrap)],
        )
        rendered = HTMLDocument(ui_res).render(lib_prefix=lib_prefix)
        self._ensure_web_dependencies(rendered["dependencies"])
        return rendered

    def _render_page_from_file(self, file: Path, lib_prefix: str) -> RenderedHTML:
        with open(file, "r") as f:
            page_html = f.read()

        doc = HTMLTextDocument(
            page_html,
            deps=[require_deps(), jquery_deps(), *shiny_deps(include_css=True)],
            deps_replace_pattern='<meta name="shiny-dependency-placeholder" content="">',
        )

        rendered = doc.render(lib_prefix=lib_prefix)
        self._ensure_web_dependencies(rendered["dependencies"])

        return rendered


def is_uifunc(x: Path | Tag | TagList | Callable[[Request], Tag | TagList]) -> bool:
    if (
        isinstance(x, Path)
        or isinstance(x, Tag)
        or isinstance(x, TagList)
        or not callable(x)
    ):
        return False
    else:
        return True


def html_dep_name(dep: HTMLDependency) -> str:
    return dep.name + "-" + str(dep.version)


def create_static_asset_route(
    mount_point: str, static_asset_path: Path
) -> starlette.routing.BaseRoute:
    """
    Create a Starlette route for serving static assets.

    Parameters
    ----------
    mount_point
        The mount point where the static assets will be served.
    static_asset_path
        The path on disk to the static assets.
    """
    if static_asset_path.is_dir():
        return starlette.routing.Mount(
            mount_point,
            StaticFiles(directory=static_asset_path),
            name="shiny-app-static-assets-" + mount_point,
        )
    else:
        mime_type = guess_mime_type(static_asset_path, strict=False)

        def file_response_handler(req: Request) -> FileResponse:
            return FileResponse(static_asset_path, media_type=mime_type)

        return starlette.routing.Route(
            mount_point,
            file_response_handler,
            name="shiny-app-static-assets-" + mount_point,
        )


def noop_server_fn(input: Inputs, output: Outputs, session: Session) -> None:
    pass


def wrap_server_fn_with_output_session(
    server: Callable[[Inputs], None]
) -> Callable[[Inputs, Outputs, Session], None]:
    def _server(input: Inputs, output: Outputs, session: Session):
        # Only has 1 parameter, ignore output, session
        server(input)

    return _server



# pyright: reportUntypedFunctionDecorator=false
# pyright: reportUnknownParameterType=false
# pyright: reportUnknownMemberType=false
# pyright: reportInvalidTypeForm=false
# pyright: reportUnknownArgumentType=false
# pyright: reportArgumentType=false
from __future__ import annotations

import warnings
from typing import Literal, overload

from shiny import App, Inputs, Outputs, Session, ui
from shiny._deprecated import ShinyDeprecationWarning
from shiny.render.transformer import (
    TransformerMetadata,
    ValueFn,
    output_transformer,
    resolve_value_fn,
)

warnings.filterwarnings("ignore", category=ShinyDeprecationWarning)

#######
# Package authors can create their own output transformer methods by leveraging
# `output_transformer` decorator.
#
# The transformer is kept simple for demonstration purposes, but it can be much more
# complex (e.g. shiny.render.plotly)
#######


@output_transformer()
async def CapitalizeTransformer(
    # Contains information about the render call: `name` and `session`
    _meta: TransformerMetadata,
    # The app-supplied output value function
    _fn: ValueFn[str | None],
    *,
    # Extra parameters that app authors can supply to the render decorator
    # (e.g. `@render_capitalize(to="upper")`)
    to: Literal["upper", "lower"] = "upper",
) -> str | None:
    # Get the value
    value = await resolve_value_fn(_fn)
    # Equvalent to:
    # if shiny.render.transformer.is_async_callable(_fn):
    #     value = await _fn()
    # else:
    #     value = _fn()

    # Render nothing if `value` is `None`
    if value is None:
        return None

    if to == "upper":
        return value.upper()
    if to == "lower":
        return value.lower()
    raise ValueError(f"Invalid value for `to`: {to}")


# First, create an overload where users can supply the extra parameters.
# Example of usage:
# ```
# @render_capitalize(to="upper")
# def value():
#     return input.caption()
# ```
# Note: Return type is `OutputRendererDecorator`
@overload
def render_capitalize(
    *,
    to: Literal["upper", "lower"] = "upper",
) -> CapitalizeTransformer.OutputRendererDecorator: ...


# Second, create an overload where users are not using parentheses to the method.
# While it doesn't look necessary, it is needed for the type checker.
# Example of usage:
# ```
# @render_capitalize
# def value():
#     return input.caption()
# ```
# Note: `_fn` type is the transformer's `ValueFn`
# Note: Return type is the transformer's `OutputRenderer`
@overload
def render_capitalize(
    _fn: CapitalizeTransformer.ValueFn,
) -> CapitalizeTransformer.OutputRenderer: ...


# Lastly, implement the renderer.
# Note: `_fn` type is the transformer's `ValueFn` or `None`
# Note: Return type is the transformer's `OutputRenderer` or `OutputRendererDecorator`
def render_capitalize(
    _fn: CapitalizeTransformer.ValueFn | None = None,
    *,
    to: Literal["upper", "lower"] = "upper",
) -> (
    CapitalizeTransformer.OutputRenderer | CapitalizeTransformer.OutputRendererDecorator
):
    return CapitalizeTransformer(
        _fn,
        CapitalizeTransformer.params(to=to),
    )


#######
# End of package author code
#######

app_ui = ui.page_fluid(
    ui.h1("Capitalization renderer"),
    ui.input_text("caption", "Caption:", "Data summary"),
    "Renderer called with out parentheses:",
    ui.output_text_verbatim("no_output", placeholder=True),
    ui.output_text_verbatim("no_parens", placeholder=True),
    "To upper:",
    ui.output_text_verbatim("to_upper", placeholder=True),
    "To lower:",
    ui.output_text_verbatim("to_lower", placeholder=True),
)


def server(input: Inputs, output: Outputs, session: Session):
    # Without parentheses
    @render_capitalize
    def no_output():
        return input.caption()

    @output
    # Without parentheses
    @render_capitalize
    def no_parens():
        return input.caption()

    # @output # Do not include to make sure auto registration works
    # With parentheses. Equivalent to `@render_capitalize()`
    @render_capitalize(to="upper")
    def to_upper():
        return input.caption()

    # provide a custom name to make sure the name can be overridden
    @output(id="to_lower")
    @render_capitalize(to="lower")
    # Works with async output value functions
    async def _():
        return input.caption()


app = App(app_ui, server)



import shiny.experimental as x
from shiny import App, ui

app_ui = ui.page_fluid(
    ui.card(
        ui.card_header("This is the header"),
        x.ui.card_title("This is the title"),
        ui.p("This is the body."),
        x.ui.card_image(
            file=None,
            src="https://posit.co/wp-content/uploads/2022/10/Posit-logo-h-full-color-RGB-TM.svg",
        ),
        ui.p("This is still the body."),
        ui.card_footer("This is the footer"),
        full_screen=True,
        id="card1",
    ),
)


app = App(app_ui, server=None)



import random

from shiny import reactive
from shiny.express import input, render, ui
from shiny.ui import output_ui

ui.markdown(
    f"""
    This example demonstrates how `@reactive.event()` can be used to restrict
    execution of: (1) a `@render` function, (2) `@reactive.calc`, or (3)
    `@reactive.effect`.

    In all three cases, the output is dependent on a random value that gets updated
    every 0.5 seconds (currently, it is {output_ui("number", inline=True)}), but
    the output is only updated when the button is clicked.
    """
)

# Always update this output when the number is updated
with ui.hold():

    @render.ui
    def number():
        return val.get()


ui.input_action_button("btn_out", "(1) Update number")


# Since ignore_none=False, the function executes before clicking the button.
# (input.btn_out() is 0 on page load, but @@reactive.event() treats 0 as None for
# action buttons.)
@render.text
@reactive.event(input.btn_out, ignore_none=False)
def out_out():
    return str(val.get())


ui.input_action_button("btn_calc", "(2) Show 1 / number")


@render.text
def out_calc():
    return str(calc())


ui.input_action_button("btn_effect", "(3) Log number")
ui.div(id="out_effect")


# Update a random number every second
val = reactive.value(random.randint(0, 1000))


@reactive.effect
def _():
    reactive.invalidate_later(0.5)
    val.set(random.randint(0, 1000))


@reactive.calc
@reactive.event(input.btn_calc)
def calc():
    return 1 / val.get()


@reactive.effect
@reactive.event(input.btn_effect)
def _():
    ui.insert_ui(
        ui.p("Random number! ", val.get()),
        selector="#out_effect",
        where="afterEnd",
    )



from shiny import App, Inputs, Outputs, Session, render, ui

app_ui = ui.page_fluid(
    ui.card(
        ui.layout_sidebar(
            ui.sidebar("Left sidebar content", id="sidebar_left"),
            ui.output_text_verbatim("state_left"),
        )
    ),
    ui.card(
        ui.layout_sidebar(
            ui.sidebar("Right sidebar content", id="sidebar_right", position="right"),
            ui.output_text_verbatim("state_right"),
        ),
    ),
    ui.card(
        ui.layout_sidebar(
            ui.sidebar("Closed sidebar content", id="sidebar_closed", open="closed"),
            ui.output_text_verbatim("state_closed"),
        )
    ),
    ui.card(
        ui.layout_sidebar(
            ui.sidebar("Always sidebar content", id="sidebar_always", open="always"),
            ui.output_text_verbatim("state_always"),
        )
    ),
)


def server(input: Inputs, output: Outputs, session: Session):
    @render.text
    def state_left():
        return f"input.sidebar_left(): {input.sidebar_left()}"

    @render.text
    def state_right():
        return f"input.sidebar_right(): {input.sidebar_right()}"

    @render.text
    def state_closed():
        return f"input.sidebar_closed(): {input.sidebar_closed()}"

    @render.text
    def state_always():
        return f"input.sidebar_always(): {input.sidebar_always()}"


app = App(app_ui, server)



from shiny import App, ui

app_ui = ui.page_fluid(
    ui.card(
        ui.card_header("This is the header"),
        ui.p("This is the body."),
        ui.p("This is still the body."),
        ui.card_footer("This is the footer"),
        full_screen=True,
    )
)


app = App(app_ui, server=None)



from shiny.express import ui

ui.page_opts(title="Nav Panel Example")

with ui.nav_panel("Page 1"):
    "Page 1 content"

with ui.nav_panel("Page 2"):
    with ui.navset_card_underline():
        with ui.nav_panel("Tab 1"):
            "Tab 1 content"
        with ui.nav_panel("Tab 2"):
            "Tab 2 content"
        with ui.nav_panel("Tab 3"):
            "Tab 3 content"



from __future__ import annotations

from typing import Literal

from playwright.sync_api import Locator, Page
from playwright.sync_api import expect as playwright_expect

from .._types import PatternOrStr, Timeout
from ..expect._internal import (
    expect_attribute_to_have_value as _expect_attribute_to_have_value,
)
from ..expect._internal import expect_class_to_have_value as _expect_class_to_have_value
from ..expect._internal import expect_style_to_have_value as _expect_style_to_have_value
from ._base import UiWithContainer


class Sidebar(
    UiWithContainer,
):
    """Controller for :func:`shiny.ui.sidebar`."""

    loc_container: Locator
    """
    Playwright `Locator` for the sidebar layout.
    """
    loc: Locator
    """
    Playwright `Locator` for the sidebar.
    """
    loc_handle: Locator
    """
    Playwright `Locator` for the open/close handle of the sidebar.
    """
    loc_position: Locator
    """
    Playwright `Locator` for the position of the sidebar.
    """
    loc_content: Locator
    """
    Playwright `Locator` for the content of the sidebar.
    """
    loc_title: Locator
    """
    Playwright `Locator` for the title of the sidebar.
    """

    def __init__(self, page: Page, id: str) -> None:
        """
        Initializes a sidebar control.

        Parameters
        ----------
        page
            Playwright `Page` of the Shiny app.
        id
            The ID of the sidebar.
        """
        super().__init__(
            page,
            id=id,
            loc=f"> aside#{id}",
            loc_container="div.bslib-sidebar-layout",
        )
        self.loc_handle = self.loc_container.locator("button.collapse-toggle")
        self.loc_position = self.loc.locator("..")
        self.loc_content = self.loc.locator("> div.sidebar-content")
        self.loc_title = self.loc_content.locator("> header.sidebar-title")

    def expect_text(self, value: PatternOrStr, *, timeout: Timeout = None) -> None:
        """
        Asserts that the sidebar has the expected text.

        Parameters
        ----------
        value
            The expected text in the sidebar.
        timeout
            The maximum time to wait for the text to appear. Defaults to `None`.
        """
        playwright_expect(self.loc_content).to_have_text(value, timeout=timeout)

    def expect_class(
        self,
        class_name: str,
        *,
        has_class: bool = True,
        timeout: Timeout = None,
    ) -> None:
        """
        Asserts that the sidebar has or does not have a CSS class.

        Parameters
        ----------
        class_name
            The CSS class to check for.
        has_class
            `True` if the sidebar should have the CSS class, `False` otherwise.
        timeout
            The maximum time to wait for the sidebar to appear. Defaults to `None`.
        """
        _expect_class_to_have_value(
            self.loc,
            class_name,
            has_class=has_class,
            timeout=timeout,
        )

    def expect_width(self, value: PatternOrStr, *, timeout: Timeout = None) -> None:
        """
        Asserts that the sidebar has the expected width.

        Parameters
        ----------
        value
            The expected width of the sidebar.
        timeout
            The maximum time to wait for the width to appear. Defaults to `None`.
        """
        _expect_style_to_have_value(
            self.loc_container, "--_sidebar-width", value, timeout=timeout
        )

    def expect_gap(self, value: PatternOrStr, *, timeout: Timeout = None) -> None:
        """
        Asserts that the sidebar has the expected gap.

        Parameters
        ----------
        value
            The expected gap of the sidebar.
        timeout
            The maximum time to wait for the gap to appear. Defaults to `None`.
        """
        _expect_style_to_have_value(self.loc_content, "gap", value, timeout=timeout)

    def expect_bg_color(self, value: PatternOrStr, *, timeout: Timeout = None) -> None:
        """
        Asserts that the sidebar has the expected background color.

        Parameters
        ----------
        value
            The expected background color of the sidebar.
        timeout
            The maximum time to wait for the background color to appear. Defaults to `None`.
        """
        _expect_style_to_have_value(
            self.loc_container, "--_sidebar-bg", value, timeout=timeout
        )

    def expect_desktop_state(
        self, value: Literal["open", "closed", "always"], *, timeout: Timeout = None
    ) -> None:
        """
        Asserts that the sidebar has the expected state on desktop.

        Parameters
        ----------
        value
            The expected state of the sidebar on desktop.
        timeout
            The maximum time to wait for the state to appear. Defaults to `None`.
        """
        _expect_attribute_to_have_value(
            self.loc_container,
            name="data-open-desktop",
            value=value,
            timeout=timeout,
        )

    def expect_mobile_state(
        self, value: Literal["open", "closed", "always"], *, timeout: Timeout = None
    ) -> None:
        """
        Asserts that the sidebar has the expected state on mobile.

        Parameters
        ----------
        value
            The expected state of the sidebar on mobile.
        timeout
            The maximum time to wait for the state to appear. Defaults to `None`.
        """
        _expect_attribute_to_have_value(
            self.loc_container,
            name="data-open-mobile",
            value=value,
            timeout=timeout,
        )

    def expect_mobile_max_height(
        self, value: PatternOrStr, *, timeout: Timeout = None
    ) -> None:
        """
        Asserts that the sidebar has the expected maximum height on mobile.

        Parameters
        ----------
        value
            The expected maximum height of the sidebar on mobile.
        timeout
            The maximum time to wait for the maximum height to appear. Defaults to `None`.
        """
        self.expect_mobile_state("always", timeout=timeout)
        _expect_style_to_have_value(
            self.loc_container, "--_mobile-max-height", value, timeout=timeout
        )

    def expect_title(self, value: PatternOrStr, *, timeout: Timeout = None) -> None:
        """
        Asserts that the sidebar has the expected title.

        Parameters
        ----------
        value
            The expected title of the sidebar.
        timeout
            The maximum time to wait for the title to appear. Defaults to `None`.
        """
        playwright_expect(self.loc_title).to_have_text(value, timeout=timeout)

    def expect_padding(
        self, value: str | list[str], *, timeout: Timeout = None
    ) -> None:
        """
        Asserts that the sidebar has the expected padding.

        Parameters
        ----------
        value
            The expected padding of the sidebar.
        timeout
            The maximum time to wait for the padding to appear. Defaults to `None`.
        """
        if not isinstance(value, list):
            value = [value]
        padding_val = " ".join(value)
        _expect_style_to_have_value(
            self.loc_content, "padding", padding_val, timeout=timeout
        )

    def expect_position(
        self,
        value: Literal["left", "right"],
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Asserts that the sidebar is in the expected position.

        Parameters
        ----------
        value
            The expected position of the sidebar.
        timeout
            The maximum time to wait for the sidebar to appear. Defaults to `None`.
        """
        is_right_sidebar = value == "right"
        _expect_class_to_have_value(
            self.loc_position,
            f"sidebar-{value}",
            has_class=is_right_sidebar,
            timeout=timeout,
        )

    def expect_handle(self, exists: bool, *, timeout: Timeout = None) -> None:
        """
        Asserts that the sidebar handle exists or does not exist.

        Parameters
        ----------
        exists
            `True` if the sidebar open/close handle should exist, `False` otherwise.
        timeout
            The maximum time to wait for the sidebar handle to appear. Defaults to `None`.
        """
        playwright_expect(self.loc_handle).to_have_count(int(exists), timeout=timeout)

    def expect_open(self, value: bool, *, timeout: Timeout = None) -> None:
        """
        Expect the sidebar to be open or closed.

        Parameters
        ----------
        value
            `True` if the sidebar should be open, `False` to be closed.
        timeout
            The maximum time to wait for the sidebar to open or close. Defaults to `None`.
        """
        playwright_expect(self.loc_handle).to_have_attribute(
            "aria-expanded", str(value).lower(), timeout=timeout
        )

    def set(self, open: bool, *, timeout: Timeout = None) -> None:
        """
        Sets the sidebar to be open or closed.

        Parameters
        ----------
        open
            `True` to open the sidebar and `False` to close it.
        timeout
            The maximum time to wait for the sidebar to open or close. Defaults to `None`.
        """
        if open ^ (self.loc_handle.get_attribute("aria-expanded") == "true"):
            self._toggle(timeout=timeout)

    def _toggle(self, *, timeout: Timeout = None) -> None:
        """
        Toggles the sidebar open or closed.

        Parameters
        ----------
        timeout
            The maximum time to wait for the sidebar to toggle. Defaults to `None`.
        """
        self.loc_handle.wait_for(state="visible", timeout=timeout)
        self.loc_handle.scroll_into_view_if_needed(timeout=timeout)
        self.loc_handle.click(timeout=timeout)



from htmltools import HTML

from shiny import App, Inputs, Outputs, Session, render, ui

app_ui = ui.page_fluid(
    ui.row(
        ui.column(
            6,
            ui.input_radio_buttons(
                "radio1",
                "Radio 1:",
                {
                    "a": HTML("<span style='color:red;'>A</span>"),
                    "b": "B",
                    "c": HTML("<span style='color:green;'>C</span>"),
                },
            ),
            ui.output_text_verbatim("radio1_out", placeholder=True),
        ),
        ui.column(
            6,
            ui.input_radio_buttons(
                "radio2",
                "Radio 2 (inline):",
                {
                    "d": HTML("<span style='color:purple;'>D</span>"),
                    "e": "E",
                    "f": HTML("<span style='color:orange;'>F</span>"),
                },
                inline=True,
            ),
            ui.output_text_verbatim("radio2_out", placeholder=True),
        ),
    ),
    ui.row(
        ui.column(
            6,
            ui.input_checkbox_group(
                "check1",
                "Check 1:",
                {
                    "red": ui.span("RED", style="color: #FF0000;"),
                    "green": ui.span("GREEN", style="color: #00AA00;"),
                    "blue": ui.span("BLUE", style="color: #0000AA;"),
                },
            ),
            ui.output_text_verbatim("check1_out", placeholder=True),
        ),
        ui.column(
            6,
            ui.input_checkbox_group(
                "check2",
                "Check 2 (inline):",
                {
                    "magenta": ui.span("MAGENTA", style="color: #FF00AA;"),
                    "orange": ui.span("ORANGE", style="color: #FFAA00;"),
                    "teal": ui.span("TEAL", style="color: #00AAAA;"),
                },
                inline=True,
            ),
            ui.output_text_verbatim("check2_out", placeholder=True),
        ),
    ),
)


def server(input: Inputs, output: Outputs, session: Session):
    @render.text
    def radio1_out():
        return input.radio1()

    @render.text
    def radio2_out():
        return input.radio2()

    @render.text
    def check1_out():
        return input.check1()

    @render.text
    def check2_out():
        return input.check2()


app = App(app_ui, server)



from shiny import render
from shiny.express import ui

chat = ui.Chat(id="chat")
chat.ui()


@chat.transform_assistant_response
def transform(content: str, chunk: str, done: bool):
    if done:
        return content + "...DONE!"
    else:
        return content


@chat.on_user_submit
async def _():
    await chat.append_message_stream(("Simple ", "response"))


"Message state:"


@render.code
def message_state():
    return str(chat.messages())



from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.run import ShinyAppProc


def test_validate_data_edit_mode(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    data_frame = controller.OutputDataFrame(page, "penguins_df")
    data_frame.expect_column_labels(
        [
            "studyName",
            "Sample Number",
            "Species",
            "Region",
            "Island",
            "Stage",
            "Individual ID",
            "Clutch Completion",
            "Date Egg",
            "Culmen Length (mm)",
            "Culmen Depth (mm)",
            "Flipper Length (mm)",
            "Body Mass (g)",
            "Sex",
            "Delta 15 N (o/oo)",
            "Delta 13 C (o/oo)",
            "Comments",
        ],
    )

    data_frame.expect_cell("PAL0708", row=0, col=0)
    data_frame.expect_ncol(17)
    data_frame.expect_class_state("ready", row=0, col=0)
    data_frame._expect_column_label(["Species"], col=3)
    data_frame._edit_cell_no_save("Study0708_edited", row=0, col=0)
    data_frame.expect_class_state("editing", row=0, col=0)
    data_frame.set_cell("Study0708_edited", row=0, col=0, finish_key="Enter")
    # data_frame.expect_class_state("saving", row=0, col=0)
    data_frame.expect_class_state("editing", row=1, col=0)
    data_frame.expect_class_state("success", row=0, col=0)
    data_frame.expect_cell("Study0708_edited", row=0, col=0)

    data_frame.expect_cell("Torgersen", row=0, col=4)
    data_frame.set_cell("Stonington", row=0, col=4, finish_key="Enter")
    data_frame.expect_class_state("failure", row=0, col=4)
    data_frame.expect_cell_title(
        "Penguin island should be one of 'Torgersen', 'Biscoe', 'Dream'", row=0, col=4
    )
    data_frame.expect_cell("Torgersen", row=0, col=4)

    data_frame.expect_cell("39.5", row=1, col=9)
    data_frame.set_cell("Stonington", row=1, col=9, finish_key="Shift+Enter")
    data_frame.expect_class_state("editing", row=0, col=9)



# pyright:basic
import time

import numpy as np
import seaborn as sns

from shiny import App, module, reactive, render, ui


# -- Reusable card module --
@module.ui
def card_ui(spinner_type):
    return ui.card(
        ui.busy_indicators.options(spinner_type=spinner_type),
        ui.card_header("Spinner: " + spinner_type),
        ui.output_plot("plot"),
    )


@module.server
def card_server(input, output, session, rerender):
    @render.plot
    def plot():
        rerender()
        time.sleep(1)
        sns.lineplot(x=np.arange(100), y=np.random.randn(100))


# -- Main app --
app_ui = ui.page_fillable(
    ui.input_task_button("rerender", "Re-render"),
    ui.layout_columns(
        card_ui("ring", "ring"),
        card_ui("bars", "bars"),
        card_ui("dots", "dots"),
        card_ui("pulse", "pulse"),
        col_widths=6,
    ),
)


def server(input, output, session):

    @reactive.calc
    def rerender():
        return input.rerender()

    card_server("ring", rerender=rerender)
    card_server("bars", rerender=rerender)
    card_server("dots", rerender=rerender)
    card_server("pulse", rerender=rerender)


app = App(app_ui, server)



from shiny import App, Inputs, Outputs, Session, reactive, ui

app_ui = ui.page_fluid(
    ui.input_action_button("add", "Add UI"),
)


def server(input: Inputs, output: Outputs, session: Session):
    @reactive.effect
    @reactive.event(input.add)
    def _():
        ui.insert_ui(
            ui.input_text("txt" + str(input.add()), "Enter some text"),
            selector="#add",
            where="afterEnd",
        )


app = App(app_ui, server)



from html import escape  # noqa: F401

from shiny.express import input, render, ui

states = {
    "East Coast": {"NY": "New York", "NJ": "New Jersey", "CT": "Connecticut"},
    "West Coast": {"WA": "Washington", "OR": "Oregon", "CA": "California"},
    "Midwest": {"MN": "Minnesota", "WI": "Wisconsin", "IA": "Iowa"},
}

ui.input_selectize(
    "state",
    "Choose a state:",
    states,
    multiple=True,
)


@render.text
def value():
    return "You choose: " + str(input.state())


ui.input_selectize(
    "state2",
    "Selectize Options",
    states,
    multiple=True,
    options=(
        {
            "placeholder": "Enter text",
            "render": ui.js_eval(
                '{option: function(item, escape) {return "<div><strong>Select " + escape(item.label) + "</strong></div>";}}'
            ),
            "create": True,
        }
    ),
)
ui.input_selectize(
    "state3",
    "Selectize plugins",
    states,
    multiple=True,
    options={"plugins": ["clear_button"]},
)



from shiny import App, Inputs, Outputs, Session, reactive, ui

items = [
    ui.accordion_panel(f"Section {letter}", f"Some narrative for section {letter}")
    for letter in "ABCDE"
]

app_ui = ui.page_fluid(
    ui.input_action_button("set_acc", "Only open sections A,C,E", class_="mt-3 mb-3"),
    # Provide an id to create a shiny input binding
    ui.accordion(*items, id="acc", open=["Section B", "Section D"], multiple=True),
)


def server(input: Inputs, output: Outputs, session: Session):
    @reactive.effect
    @reactive.event(input.set_acc)
    def _():
        ui.update_accordion("acc", show=["Section A", "Section C", "Section E"])


app = App(app_ui, server)



from shiny import reactive
from shiny.express import input, render, ui

ui.input_action_button("add", "Add more controls")


@render.ui
@reactive.event(input.add)
def moreControls():
    return [
        ui.input_slider("n", "N", min=1, max=1000, value=500),
        ui.input_text("label", "Label"),
    ]



from __future__ import annotations

from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.run import ShinyAppProc


def test_update_data(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    df = controller.OutputDataFrame(page, "df")
    df_selected = controller.OutputDataFrame(page, "df_selected")
    shift_btn = controller.InputActionButton(page, "shift_btn")
    different_btn = controller.InputActionButton(page, "different_btn")

    df.expect_nrow(344)
    df.expect_ncol(2)
    df_selected.expect_column_labels(["studyName", "Sample Number"])

    df_selected.expect_nrow(0)
    df_selected.expect_ncol(2)
    df_selected.expect_column_labels(["studyName", "Sample Number"])

    df.select_rows([1])
    df_selected.expect_nrow(1)
    df_selected.expect_cell("2", row=0, col=1)

    # Shift data
    shift_btn.click()
    df.expect_nrow(2)
    df.expect_ncol(2)
    df.expect_column_labels(["studyName", "Sample Number"])
    df.expect_selected_rows([1])
    df.expect_cell("3", row=0, col=1)
    df.expect_cell("4", row=1, col=1)
    df_selected.expect_nrow(1)
    df_selected.expect_ncol(2)
    df_selected.expect_cell("4", row=0, col=1)

    # Change data set
    different_btn.click()
    df.expect_nrow(26)
    df.expect_ncol(2)
    df.expect_column_labels(["Letter", "Negative index"])
    df.expect_selected_rows([1])

    df_selected.expect_nrow(1)
    df_selected.expect_ncol(2)
    df_selected.expect_cell("b", row=0, col=0)
    df_selected.expect_cell("-2", row=0, col=1)

    # Change data set again
    different_btn.click()
    df.expect_nrow(344)
    df.expect_ncol(17)
    df.expect_column_labels(
        [
            "studyName",
            "Sample Number",
            "Species",
            "Region",
            "Island",
            "Stage",
            "Individual ID",
            "Clutch Completion",
            "Date Egg",
            "Culmen Length (mm)",
            "Culmen Depth (mm)",
            "Flipper Length (mm)",
            "Body Mass (g)",
            "Sex",
            "Delta 15 N (o/oo)",
            "Delta 13 C (o/oo)",
            "Comments",
        ]
    )
    df.expect_selected_rows([1])



# flake8: noqa E402
# This file is used to create a minimal Bootstrap CSS file based on a custom Bootstrap
# theme created with ui.Theme*().

# NOTE: This script requires the `purgecss` package to be installed.
#       You can install it with `npm install -g purgecss`.

import shutil
import subprocess
from pathlib import Path

from shared import filler_text, my_theme

my_theme.name = "demo"

from shiny import ui

# If __file__ is not defined, use the current working directory
if not globals().get("__file__"):
    __file__ = Path.cwd() / "_purgecss.py"


app_ui = ui.page_sidebar(
    ui.sidebar(
        ui.input_numeric("n", "N", min=0, max=100, value=20),
        title="Parameters",
    ),
    ui.h2("Output"),
    ui.output_text_verbatim("txt"),
    ui.markdown(filler_text),
    title="Theme Example",
    theme=my_theme,
)


save_dir = Path(__file__).parent / "output"
if save_dir.exists():
    shutil.rmtree(save_dir)
save_dir.mkdir()
app_ui.save_html(save_dir / "index.html", include_version=False)

purged_dir = Path(__file__).parent / "css"
purged_dir.mkdir(exist_ok=True)

args = [
    "purgecss",
    "--variables",
    "--css",
    "output/lib/shiny-theme-demo/shiny-theme-demo.min.css",
    "--content",
    "output/index.html",
    "--output",
    "css",
]

subprocess.run(args)

(purged_dir / "shiny-theme-demo.min.css").rename(purged_dir / "shiny-theme-demo.css")

if True:
    shutil.rmtree(save_dir)



from shiny.express import input, render, ui

ui.page_opts(title="Kitchen Sink for input_switch()", fillable=True)

with ui.layout_columns():
    with ui.card():
        ui.h3("Default Switch")
        ui.input_switch("default", "Default switch")

        @render.code
        def default_txt():
            return str(input.default())

    with ui.card():
        ui.h3("With Value")
        ui.input_switch("value", "Preset value", value=True)

        @render.code
        def value_txt():
            return str(input.value())

    with ui.card():
        ui.h3("Custom Width")
        ui.input_switch("width", "Custom width", width="200px")

        @render.code
        def width_txt():
            return str(input.width())



# This file is necessary for pytest to find relative module files
# such as examples/example_apps.py
from __future__ import annotations

from pathlib import PurePath

import pytest
from playwright.sync_api import BrowserContext, Page

from shiny.pytest import ScopeName as ScopeName
from shiny.pytest import create_app_fixture

__all__ = (
    "create_doc_example_fixture",
    "create_example_fixture",
    "create_doc_example_core_fixture",
    "create_doc_example_express_fixture",
)


here = PurePath(__file__).parent
here_root = here.parent.parent


# Make a single page fixture that can be used by all tests
@pytest.fixture(scope="session")
# By using a single page, the browser is only launched once and all tests run in the same tab / page.
def session_page(browser: BrowserContext) -> Page:
    """
    Create a new page within the given browser context.
    Parameters:
        browser (BrowserContext): The browser context in which to create the new page.
    Returns:
        Page: The newly created page.
    """
    return browser.new_page()


@pytest.fixture(scope="function")
# By going to `about:blank`, we _reset_ the page to a known state before each test.
# It is not perfect, but it is faster than making a new page for each test.
# This must be done before each test
def page(session_page: Page) -> Page:
    """
    Reset the given page to a known state before each test.
    The page is built on the session_page, which is maintained over the full session.
    The page will visit "about:blank" to reset between apps.
    The default viewport size is set to 1920 x 1080 (1080p) for each test function.
    Parameters:
        session_page (Page): The page to reset.
    """
    session_page.goto("about:blank")
    # Reset screen size to 1080p
    session_page.set_viewport_size({"width": 1920, "height": 1080})
    return session_page


def create_example_fixture(
    example_name: str,
    example_file: str = "app.py",
    scope: ScopeName = "module",
):
    """Used to create app fixtures from apps in py-shiny/examples"""
    return create_app_fixture(
        here_root / "examples" / example_name / example_file, scope
    )


def create_doc_example_fixture(
    example_name: str,
    example_file: str = "app.py",
    scope: ScopeName = "module",
):
    """Used to create app fixtures from apps in py-shiny/shiny/api-examples"""
    return create_app_fixture(
        here_root / "shiny/api-examples" / example_name / example_file, scope
    )


def create_doc_example_core_fixture(
    example_name: str,
    scope: ScopeName = "module",
):
    """Used to create app fixtures from ``app-core.py`` example apps in py-shiny/shiny/api-examples"""
    return create_doc_example_fixture(example_name, "app-core.py", scope)


def create_doc_example_express_fixture(
    example_name: str,
    scope: ScopeName = "module",
):
    """Used to create app fixtures from ``app-express.py`` example apps in py-shiny/shiny/api-examples"""
    return create_doc_example_fixture(example_name, "app-express.py", scope)



from __future__ import annotations

import glob
from pathlib import Path
from typing import Dict, Set

from tests.pytest._utils import skip_on_windows

known_entries: Dict[str, Set[str]] = {
    # "tests/pytest/test_poll.py": {
    #     "my_locator.filter('foo')",
    # }
}

# Trim all line values of `known_entries`
for k, v in known_entries.items():
    known_entries[k] = {x.strip() for x in v}


@skip_on_windows
def test_named_temporary_file_is_not_used():
    """
    Playwright's Locator hangs when we call `.filter(foo)`.
    Instead you should use `.locator("xpath=.", has=page.locator(foo))`
    """

    root_here = Path(__file__).parent.parent.parent
    shiny_files = glob.glob(str(root_here / "shiny" / "**" / "*.py"), recursive=True)
    tests_files = glob.glob(str(root_here / "tests" / "**" / "*.py"), recursive=True)

    assert len(shiny_files) > 0
    assert len(tests_files) > 0

    all_files = [*shiny_files, *tests_files]

    search_string = ".filter("

    bad_entries: list[tuple[Path, int, str]] = []

    # For every python file...
    for path in all_files:
        path = Path(path)
        # Skip if dir
        if path.is_dir():
            continue

        # Skip this file
        if path.name in {"test_playwright_filter.py"}:
            continue

        with open(path, "r") as f:
            # Read file contents
            txt = f.read().replace(".filter()", ".not_playwright_filter()")

            # Skip if search string is not in file
            if search_string not in txt:
                continue

            # Split file contents by line
            lines = txt.split("\n")
            rel_path = path.relative_to(root_here)
            known_lines = known_entries.get(str(rel_path), set())
            seen_lines: set[str] = set()

            # If the search string is in the line
            # and the line is not in the known lines,
            # add it to the bad entries
            for i, line in enumerate(lines):
                line = line.strip()
                if search_string in line:
                    seen_lines.add(line)
                    if line not in known_lines:
                        bad_entries.append((rel_path, i + 1, line))

            if (len(known_lines) > 0) and (len(seen_lines) != len(known_lines)):
                raise ValueError(
                    f"Lines not found in {rel_path}: {known_lines - seen_lines}"
                    "\nPlease remove them from the known_entries dictionary."
                )

    assert (
        len(bad_entries) == 0
    ), f"Unexpected files containing `.filter(`: {str(bad_entries)}"



from typing import Dict, Tuple

import matplotlib.colors as mpl_colors
import seaborn as sns  # type: ignore

# "darkorange", "purple", "cyan4"
colors = [[255, 140, 0], [160, 32, 240], [0, 139, 139]]
colors = [(r / 255.0, g / 255.0, b / 255.0) for r, g, b in colors]

palette: Dict[str, Tuple[float, float, float]] = {
    "Adelie": colors[0],
    "Chinstrap": colors[1],
    "Gentoo": colors[2],
    "default": sns.color_palette()[0],  # type: ignore
}

bg_palette = {}
# bgcols: list[str] = sns.color_palette().as_hex()
# Use `sns.set_style("whitegrid")` to help find approx alpha value
for name, col in palette.items():
    # Adjusted n_colors until `axe` accessibility did not complain about color contrast
    bg_palette[name] = mpl_colors.to_hex(sns.light_palette(col, n_colors=7)[1])  # type: ignore



import matplotlib.pyplot as plt

from shiny import App, Inputs, Outputs, Session, render, ui

app_ui = ui.page_fluid(
    ui.input_checkbox("render", "Render", value=True),
    ui.panel_conditional(
        "input.render",
        ui.tags.h5("A plot should appear immediately below this text."),
    ),
    ui.output_plot("mpl"),
    ui.tags.hr(),
    ui.panel_conditional(
        "input.render",
        ui.tags.h5("An error message should appear immediately below this text."),
    ),
    ui.output_plot("mpl_bad"),
)


def server(input: Inputs, output: Outputs, session: Session):
    @render.plot
    def mpl():
        if input.render():
            plt.hist([1, 1, 2, 3, 5])

    @render.plot
    async def mpl_bad():
        if input.render():
            plt.hist([1, 1, 2, 3, 5])


app = App(app_ui, server)



from ._transformer import (  # noqa: F401
    TransformerMetadata,
    TransformerParams,
    OutputRenderer,
    output_transformer,
    is_async_callable,
    ValueFn,
    ValueFnSync,  # pyright: ignore[reportUnusedImport]
    ValueFnAsync,  # pyright: ignore[reportUnusedImport]
    TransformFn,  # pyright: ignore[reportUnusedImport]
    OutputTransformer,  # pyright: ignore[reportUnusedImport]
    OutputRendererSync,  # pyright: ignore[reportUnusedImport]
    OutputRendererAsync,  # pyright: ignore[reportUnusedImport]
    resolve_value_fn,  # pyright: ignore[reportUnusedImport]
)

__all__ = (
    "TransformerMetadata",
    "TransformerParams",
    "OutputRenderer",
    "ValueFn",
    "output_transformer",
    "is_async_callable",
)



from __future__ import annotations

import re
from contextlib import contextmanager
from contextvars import ContextVar, Token
from typing import Generator, Pattern, Union, overload


class ResolvedId(str):
    def __call__(self, id: Id) -> ResolvedId:
        if isinstance(id, ResolvedId):
            return id

        validate_id(id)

        if self == "":
            return ResolvedId(id)
        else:
            return ResolvedId(self + "-" + id)


Root: ResolvedId = ResolvedId("")


Id = Union[str, ResolvedId]


def current_namespace() -> ResolvedId:
    return _current_namespace.get() or _default_namespace


def resolve_id(id: Id) -> ResolvedId:
    """
    Resolve an ID, possibly with a module namespace.

    Parameters
    ----------
    Args
        id: An ID.

    Returns
        An ID (if in a module, this will contain a namespace prefix).
    """
    curr_ns = current_namespace()
    return curr_ns(id)


@overload
def resolve_id_or_none(id: None) -> None: ...


@overload
def resolve_id_or_none(id: Id) -> ResolvedId: ...


# Do not export this method from `shiny`. Let developers handle it themselves.
def resolve_id_or_none(id: Id | None) -> ResolvedId | None:
    """
    Resolve an ID, possibly with a module namespace. With `None` support.

    This method should only be used if `None` values are allowed. If not, use `resolve_id(id=)`.

    Parameters
    ----------
    Args
        id: An ID or `None`.

    Returns
        If `id=None`, then `None`. Otherwise, an ID (if in a module, this will contain a namespace prefix).
    """
    if id is None:
        return None
    return resolve_id(id)


# \w is a large set for unicode patterns, that's fine; we mostly want to avoid some
# special characters like space, comma, period, and especially dash
re_valid_id: Pattern[str] = re.compile("^\\.?\\w+$")


def validate_id(id: str) -> None:
    if not isinstance(id, str):
        raise ValueError("`id` must be a single string")
    if id == "":
        raise ValueError("`id` must be a non-empty string")
    if not re_valid_id.match(id):
        raise ValueError(
            f"The string '{id}' is not a valid id; only letters, numbers, and "
            "underscore are permitted"
        )


_current_namespace: ContextVar[ResolvedId | None] = ContextVar(
    "current_namespace", default=None
)
_default_namespace: ResolvedId = Root


@contextmanager
def namespace_context(id: Id | None) -> Generator[None, None, None]:
    namespace = resolve_id(id) if id else Root
    token: Token[ResolvedId | None] = _current_namespace.set(namespace)
    try:
        yield
    finally:
        _current_namespace.reset(token)



# TODO-barret: ts code to stringify objects?

from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import Any, Union, cast

import htmltools
import narwhals.stable.v1 as nw
import pandas as pd
import polars as pl
import polars.testing as pl_testing
import pytest

from shiny.render._data_frame_utils._tbl_data import (
    as_data_frame,
    serialize_dtype,
    serialize_frame,
    subset_frame,
)
from shiny.render._data_frame_utils._types import IntoDataFrame
from shiny.session import Session, session_context
from shiny.session._session import RenderedDeps, ResolvedId, Root
from shiny.ui import HTML, TagChild, TagList, h1, span


class _MockSession:
    ns: ResolvedId = Root

    # Simplified version of `AppSession._process_ui()`
    def _process_ui(self, ui: TagChild) -> RenderedDeps:
        res = TagList(ui).render()
        deps: list[dict[str, Any]] = []
        for dep in res["dependencies"]:
            # self.app._register_web_dependency(dep)
            dep_dict = dep.as_dict()
            deps.append(dep_dict)

        return {"deps": deps, "html": res["html"]}


test_session = cast(Session, _MockSession())


class C:
    x: int

    def __init__(self, x: int):
        self.x = x

    def __str__(self):
        return f"<{self.__class__.__name__} object>"


@dataclass
class D:
    y: int


html_dep = htmltools.HTMLDependency("test-dep", version="1", head="head-content")
ex_html_dep_dict = html_dep.as_dict()


DATA = {
    "num": [1, 2],
    "chr": ["a", "b"],
    "cat": ["a", "a"],
    "bool": [True, False],
    "dt": [datetime(2000, 1, 2)] * 2,
    "duration": [timedelta(weeks=1), timedelta(days=7)],
    "html": [span("span content", html_dep)] * 2,
    "html_str": [HTML("<strong>bolded</strong>")] * 2,
    "struct": [{"x": 1}, {"x": 2}],
    "arr": [[1, 2], [3, 4]],
    "object": [C(1), D(2)],
}

# Polars.Series is not always a TagNode (as it has a `__repr_html__` method)
# So we need to check if it is a TagNode to determine if it is an `"html"` or `"unknown"` type
polars_series_col_type = (
    "html" if htmltools.is_tag_node(pl.Series([[1, 2], [3, 4]])) else "unknown"
)


def pandas_dict_to_narwhals(d: dict[str, Any]) -> nw.DataFrame[pd.DataFrame]:
    return nw.from_native(pd.DataFrame(d), eager_only=True)


def polars_dict_to_narwhals(d: dict[str, Any]) -> nw.DataFrame[pl.DataFrame]:
    return nw.from_native(pl.DataFrame(d), eager_only=True)


def series_to_narwhals(ser: pd.Series[Any] | pl.Series) -> nw.Series:
    return nw.from_native(ser, series_only=True, strict=True)


params_frames = [
    pytest.param(pd.DataFrame, id="pandas"),
    pytest.param(pl.DataFrame, id="polars"),
    pytest.param(pandas_dict_to_narwhals, id="narwhals-pandas"),
    pytest.param(polars_dict_to_narwhals, id="narwhals-polars"),
]


@pytest.fixture(params=params_frames, scope="function")
def df_f(request: pytest.FixtureRequest) -> IntoDataFrame:
    return request.param(DATA)


@pytest.fixture(params=params_frames, scope="function")
def small_df_f(request: pytest.FixtureRequest) -> IntoDataFrame:
    return request.param({"x": [1, 2], "y": [3, 4]})


def assert_frame_equal(
    src: pd.DataFrame | pl.DataFrame,
    target: pd.DataFrame | pl.DataFrame,
    use_index: bool = False,
):
    if isinstance(src, pd.DataFrame):
        assert isinstance(target, pd.DataFrame)
        if use_index:
            pd.testing.assert_frame_equal(src, target)
        else:
            pd.testing.assert_frame_equal(
                src.reset_index(drop=True),
                target.reset_index(drop=True),
            )
    elif isinstance(src, pl.DataFrame):
        assert isinstance(target, pl.DataFrame)
        pl_testing.assert_frame_equal(src, target)
    else:
        raise NotImplementedError(f"Unsupported data type: {type(src)}")


def assert_frame_equal2(
    src: IntoDataFrame,
    target_dict: dict[str, Any],
    use_index: bool = False,
):
    src_native = nw.to_native(src, strict=False)
    target_native = nw.to_native(src, strict=False).__class__(target_dict)

    assert_frame_equal(src_native, target_native, use_index)


@pytest.mark.parametrize(
    "ser, res_type",
    [  # pyright: ignore[reportUnknownArgumentType] # We are explicitly setting some values to `unkown`
        # polars ----
        (pl.Series([1]), "numeric"),
        (pl.Series([1.1]), "numeric"),
        (pl.Series(["a"]), "string"),
        (pl.Series([True, False]), "boolean"),
        (pl.Series([datetime.now()]), "datetime"),
        (pl.Series([timedelta(weeks=1)]), "duration"),
        (
            pl.Series(["a", "b", "b", "c"], dtype=pl.Categorical),
            ("categorical", ["a", "b", "c"]),
        ),
        (
            pl.Series(
                ["Panda", "Polar", "Brown", "Brown", "Polar"],
                dtype=pl.Enum(["Polar", "Panda", "Brown"]),
            ),
            ("categorical", ["Polar", "Panda", "Brown"]),
        ),
        (pl.Series([{"x": 1}]), "object"),
        (pl.Series([h1("yo")]), "html"),
        (pl.Series([HTML("yo")]), "html"),
        # pandas ----
        (pd.Series([1]), "numeric"),
        (pd.Series([1.1]), "numeric"),
        (pd.Series(["a"]), "string"),
        (pd.Series([True, False]), "boolean"),
        (pd.Series([datetime.now()]), "datetime"),
        (pd.Series([timedelta(weeks=1)]), "duration"),
        (
            pd.Series(["a", "b", "b", "c"], dtype="category"),
            ("categorical", ["a", "b", "c"]),
        ),
        (
            pd.Series(
                pd.Categorical(
                    ["Panda", "Polar", "Brown", "Brown", "Polar"],
                    categories=["Polar", "Panda", "Brown"],
                )
            ),
            ("categorical", ["Polar", "Panda", "Brown"]),
        ),
        (
            pd.Series(
                pd.CategoricalIndex(
                    ["Panda", "Polar", "Brown", "Brown", "Polar"],
                )
            ),
            ("categorical", ["Brown", "Panda", "Polar"]),
        ),
        (
            pd.Series(
                pd.CategoricalIndex(
                    ["Panda", "Polar", "Brown", "Brown", "Polar"],
                    categories=["Polar", "Panda", "Brown"],
                )
            ),
            ("categorical", ["Polar", "Panda", "Brown"]),
        ),
        (pd.Series([{"x": 1}]), "object"),
        (pd.Series([h1("yo")]), "html"),
        (pd.Series([HTML("yo")]), "html"),
    ],
)
def test_serialize_dtype(
    ser: Union[
        "pd.Series[Any]",
        pl.Series,
    ],
    res_type: str | tuple[str, list[str]],
):
    nw_ser = series_to_narwhals(ser)
    dtype_info = serialize_dtype(nw_ser)
    ex_type = res_type if isinstance(res_type, str) else res_type[0]
    assert dtype_info["type"] == ex_type
    if dtype_info["type"] == "categorical":
        assert isinstance(res_type, tuple)
        assert dtype_info["categories"] == res_type[1]


def test_serialize_frame(df_f: IntoDataFrame):

    # if not isinstance(df_f, pl.DataFrame):
    #     pytest.skip()

    df_nw = as_data_frame(df_f)

    is_polars_backed = isinstance(
        nw.to_native(nw.from_native(df_f, eager_only=True)), pl.DataFrame
    )

    with session_context(test_session):
        res = serialize_frame(df_nw)
    assert res == {
        "columns": [
            "num",
            "chr",
            "cat",
            "bool",
            "dt",
            "duration",
            "html",
            "html_str",
            "struct",
            "arr",
            "object",
        ],
        "data": [
            [
                1,
                "a",
                "a",
                True,
                "2000-01-02T00:00:00" if is_polars_backed else "2000-01-02 00:00:00",
                "7 days, 0:00:00" if is_polars_backed else "7 days 00:00:00",
                {
                    "isShinyHtml": True,
                    "obj": {"deps": [], "html": "<span>span content</span>"},
                },
                {
                    "isShinyHtml": True,
                    "obj": {"deps": [], "html": "<strong>bolded</strong>"},
                },
                {"x": 1},
                [1, 2],
                "<C object>",
            ],
            [
                2,
                "b",
                "a",
                False,
                "2000-01-02T00:00:00" if is_polars_backed else "2000-01-02 00:00:00",
                "7 days, 0:00:00" if is_polars_backed else "7 days 00:00:00",
                {
                    "isShinyHtml": True,
                    "obj": {"deps": [], "html": "<span>span content</span>"},
                },
                {
                    "isShinyHtml": True,
                    "obj": {"deps": [], "html": "<strong>bolded</strong>"},
                },
                {"x": 2},
                [3, 4],
                {"y": 2},
            ],
        ],
        "typeHints": [
            {"type": "numeric"},
            {"type": "string"},
            {"type": "string"},
            {"type": "boolean"},
            {"type": "datetime"},
            {"type": "duration"},
            {"type": "html"},
            {"type": "html"},
            {"type": "object"},
            {"type": "object"},
            {"type": "object"},
        ],
        "htmlDeps": [ex_html_dep_dict],
    }


def test_subset_frame(df_f: IntoDataFrame):
    # TODO: this assumes subset_frame doesn't reset index
    res = subset_frame(as_data_frame(df_f), rows=[1], cols=["chr", "num"])
    dst = {"chr": ["b"], "num": [2]}

    assert_frame_equal2(res, dst)


def test_subset_frame_rows_single(small_df_f: IntoDataFrame):
    res = subset_frame(as_data_frame(small_df_f), rows=[1])

    assert_frame_equal2(
        res,
        {"x": [2], "y": [4]},
    )


def test_subset_frame_cols_single(small_df_f: IntoDataFrame):
    # TODO: include test of polars
    res = subset_frame(as_data_frame(small_df_f), cols=["y"])

    assert_frame_equal2(
        res,
        {"y": [3, 4]},
    )


def test_dtype_coverage():
    from pathlib import Path

    from narwhals.stable.v1 import dtypes as nw_dtypes

    # Copy from https://github.com/narwhals-dev/narwhals/blob/2c9e2e7a308ebb30c6f672e27c1da2086ebbecbc/utils/check_api_reference.py#L144-L146
    dtype_names = [
        i
        for i in cast(str, nw_dtypes.__dir__())  # pyright: ignore
        if i[0].isupper() and not i.isupper() and i[0] != "_"
    ]

    with open(
        Path(__file__).parent.parent.parent  # Repo root
        / "shiny"
        / "render"
        / "_data_frame_utils"
        / "_tbl_data.py"
    ) as f:
        tbl_data_lines = f.readlines()

    tbl_data_lines = [line for line in tbl_data_lines if "nw." in line]

    errs: list[str] = []

    for dtype_name in dtype_names:

        # Skip known types or imports that are not dtypes
        if dtype_name in (
            # narwhals
            "DType",
            "NumericType",
            "TemporalType",
            "Unknown",
            # typing import
            "Literal",
        ):
            continue

        dtype_cls = getattr(nw_dtypes, dtype_name)
        if not issubclass(dtype_cls, nw_dtypes.DType):
            continue

        if dtype_cls.is_numeric():
            continue

        if f"nw.{dtype_name}" in "".join(tbl_data_lines):
            continue

        errs.append(f"Missing: {dtype_name}")

    assert not errs, "Missing narwhals dtype implementations:\n" + "\n".join(errs)



from shiny.express import input, render, ui
from shiny.types import SilentCancelOutputException

ui.input_text(
    "txt",
    "Delete the input text completely: it won't get removed below the input",
    "Some text",
    width="400px",
)


@render.ui
def txt_out():
    if not input.txt():
        raise SilentCancelOutputException()
    return "Your input: " + input.txt()
